
from PyQt5.QtWidgets import QMainWindow, QApplication, QMessageBox, QScrollArea, QComboBox, QFrame, QCheckBox, \
    QGridLayout, QLineEdit, QVBoxLayout, QWidget, QLabel, QHBoxLayout, QPushButton
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QIcon, QDoubleValidator, QIntValidator

from celldetective.gui.gui_utils import center_window, FeatureChoice, ListWidget, QHSeperationLine, FigureCanvas, \
    GeometryChoice, OperationChoice
from superqt import QLabeledDoubleSlider
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6

from celldetective.gui.thresholds_gui import ThresholdSpot
from celldetective.utils import extract_experiment_channels, get_software_location
from celldetective.io import load_frames, auto_load_number_of_frames
from celldetective.measure import compute_haralick_features
import numpy as np
from tifffile import imread
import json
import os
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from glob import glob
from natsort import natsorted
from tifffile import imread
from pathlib import Path
import gc

from celldetective.gui.viewers import CellEdgeVisualizer
from celldetective.gui.layouts import ProtocolDesignerLayout, BackgroundFitCorrectionLayout, LocalCorrectionLayout
from celldetective.gui.gui_utils import ThresholdLineEdit
from celldetective.gui import Styles

class ConfigMeasurements(QMainWindow, Styles):
    """
    UI to set measurement instructions.

    """

    def __init__(self, parent_window=None):

        super().__init__()

        self.parent_window = parent_window
        self.setWindowTitle("Configure measurements")
        self.setWindowIcon(QIcon(os.sep.join(['celldetective', 'icons', 'mexican-hat.png'])))
        self.mode = self.parent_window.mode
        self.exp_dir = self.parent_window.exp_dir
        self.background_correction = []
        if self.mode == "targets":
            self.config_name = "btrack_config_targets.json"
            self.measure_instructions_path = self.parent_window.exp_dir + "configs/measurement_instructions_targets.json"
        elif self.mode == "effectors":
            self.config_name = "btrack_config_effectors.json"
            self.measure_instructions_path = self.parent_window.exp_dir + "configs/measurement_instructions_effectors.json"
        self.soft_path = get_software_location()
        self.clear_previous = False

        exp_config = self.exp_dir + "config.ini"
        self.config_path = self.exp_dir + self.config_name
        self.channel_names, self.channels = extract_experiment_channels(exp_config)
        self.channel_names = np.array(self.channel_names)
        self.channels = np.array(self.channels)

        self.screen_height = self.parent_window.parent_window.parent_window.screen_height
        center_window(self)

        self.onlyFloat = QDoubleValidator()
        self.onlyInt = QIntValidator()

        self.setMinimumWidth(500)
        self.setMinimumHeight(int(0.3 * self.screen_height))
        self.setMaximumHeight(int(0.8 * self.screen_height))
        self.populate_widget()
        self.load_previous_measurement_instructions()

    def populate_widget(self):

        """
        Create the multibox design.

        """

        # Create button widget and layout
        self.scroll_area = QScrollArea(self)
        self.button_widget = QWidget()
        main_layout = QVBoxLayout()
        self.button_widget.setLayout(main_layout)
        main_layout.setContentsMargins(30, 30, 30, 30)

        self.normalisation_frame = QFrame()
        self.normalisation_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        
        self.local_correction_layout = LocalCorrectionLayout(self)
        self.fit_correction_layout = BackgroundFitCorrectionLayout(self)
        
        self.protocol_layout = ProtocolDesignerLayout(parent_window=self,
                                                      tab_layouts=[ self.local_correction_layout, self.fit_correction_layout],
                                                      tab_names=['Local', 'Fit'],
                                                      title='BACKGROUND CORRECTION',
                                                      list_title='Corrections to apply:'
                                                      )
        
        self.normalisation_frame.setLayout(self.protocol_layout)

        #self.populate_normalisation_tabs()
        main_layout.addWidget(self.normalisation_frame)

        # first frame for FEATURES
        self.features_frame = QFrame()
        self.features_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        self.populate_features_frame()
        main_layout.addWidget(self.features_frame)

        # second frame for ISOTROPIC MEASUREMENTS
        self.iso_frame = QFrame()
        self.iso_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        self.populate_iso_frame()
        main_layout.addWidget(self.iso_frame)

        self.spot_detection_frame = QFrame()
        self.spot_detection_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        self.populate_spot_detection()
        main_layout.addWidget(self.spot_detection_frame)

        self.clear_previous_btn = QCheckBox('clear previous measurements')
        main_layout.addWidget(self.clear_previous_btn)

        self.submit_btn = QPushButton('Save')
        self.submit_btn.setStyleSheet(self.button_style_sheet)
        self.submit_btn.clicked.connect(self.write_instructions)
        main_layout.addWidget(self.submit_btn)

        # self.populate_left_panel()
        # grid.addLayout(self.left_side, 0, 0, 1, 1)
        self.button_widget.adjustSize()

        self.scroll_area.setAlignment(Qt.AlignCenter)
        self.scroll_area.setWidget(self.button_widget)
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.scroll_area.setWidgetResizable(True)
        self.setCentralWidget(self.scroll_area)
        self.show()

        QApplication.processEvents()
        self.adjustScrollArea()

    def populate_iso_frame(self):

        """
        Add widgets and layout in the POST-PROCESSING frame.
        """

        grid = QGridLayout(self.iso_frame)

        self.iso_lbl = QLabel("ISOTROPIC MEASUREMENTS")
        self.iso_lbl.setStyleSheet("""
            font-weight: bold;
            padding: 0px;
            """)
        grid.addWidget(self.iso_lbl, 0, 0, 1, 4, alignment=Qt.AlignCenter)
        self.generate_iso_contents()
        grid.addWidget(self.ContentsIso, 1, 0, 1, 4, alignment=Qt.AlignTop)

    def populate_features_frame(self):

        """
        Add widgets and layout in the FEATURES frame.
        """

        grid = QGridLayout(self.features_frame)

        self.feature_lbl = QLabel("FEATURES")
        self.feature_lbl.setStyleSheet("""
            font-weight: bold;
            padding: 0px;
            """)
        grid.addWidget(self.feature_lbl, 0, 0, 1, 4, alignment=Qt.AlignCenter)

        self.generate_feature_panel_contents()
        grid.addWidget(self.ContentsFeatures, 1, 0, 1, 4, alignment=Qt.AlignTop)

    def generate_iso_contents(self):

        self.ContentsIso = QFrame()
        layout = QVBoxLayout(self.ContentsIso)
        layout.setContentsMargins(0, 0, 0, 0)

        radii_layout = QHBoxLayout()
        self.radii_lbl = QLabel('Measurement radii (from center):')
        self.radii_lbl.setToolTip('Define radii or donughts for intensity measurements.')
        radii_layout.addWidget(self.radii_lbl, 90)

        self.del_radius_btn = QPushButton("")
        self.del_radius_btn.setStyleSheet(self.button_select_all)
        self.del_radius_btn.setIcon(icon(MDI6.trash_can, color="black"))
        self.del_radius_btn.setToolTip("Remove radius")
        self.del_radius_btn.setIconSize(QSize(20, 20))
        radii_layout.addWidget(self.del_radius_btn, 5)

        self.add_radius_btn = QPushButton("")
        self.add_radius_btn.setStyleSheet(self.button_select_all)
        self.add_radius_btn.setIcon(icon(MDI6.plus, color="black"))
        self.add_radius_btn.setToolTip("Add radius")
        self.add_radius_btn.setIconSize(QSize(20, 20))
        radii_layout.addWidget(self.add_radius_btn, 5)
        layout.addLayout(radii_layout)

        self.radii_list = ListWidget(GeometryChoice, initial_features=["10"], dtype=int)
        layout.addWidget(self.radii_list)

        self.del_radius_btn.clicked.connect(self.radii_list.removeSel)
        self.add_radius_btn.clicked.connect(self.radii_list.addItem)

        # Operation
        operation_layout = QHBoxLayout()
        self.op_lbl = QLabel('Operation to perform:')
        self.op_lbl.setToolTip('Set the operations to perform inside the ROI.')
        operation_layout.addWidget(self.op_lbl, 90)

        self.del_op_btn = QPushButton("")
        self.del_op_btn.setStyleSheet(self.button_select_all)
        self.del_op_btn.setIcon(icon(MDI6.trash_can, color="black"))
        self.del_op_btn.setToolTip("Remove operation")
        self.del_op_btn.setIconSize(QSize(20, 20))
        operation_layout.addWidget(self.del_op_btn, 5)

        self.add_op_btn = QPushButton("")
        self.add_op_btn.setStyleSheet(self.button_select_all)
        self.add_op_btn.setIcon(icon(MDI6.plus, color="black"))
        self.add_op_btn.setToolTip("Add operation")
        self.add_op_btn.setIconSize(QSize(20, 20))
        operation_layout.addWidget(self.add_op_btn, 5)
        layout.addLayout(operation_layout)

        self.operations_list = ListWidget(OperationChoice, initial_features=["mean"])
        layout.addWidget(self.operations_list)

        self.del_op_btn.clicked.connect(self.operations_list.removeSel)
        self.add_op_btn.clicked.connect(self.operations_list.addItem)

    def generate_feature_panel_contents(self):

        self.ContentsFeatures = QFrame()
        layout = QVBoxLayout(self.ContentsFeatures)
        layout.setContentsMargins(0, 0, 0, 0)

        feature_layout = QHBoxLayout()
        feature_layout.setContentsMargins(0, 0, 0, 0)

        self.feature_lbl = QLabel("Add features:")
        self.del_feature_btn = QPushButton("")
        self.del_feature_btn.setStyleSheet(self.button_select_all)
        self.del_feature_btn.setIcon(icon(MDI6.trash_can, color="black"))
        self.del_feature_btn.setToolTip("Remove feature")
        self.del_feature_btn.setIconSize(QSize(20, 20))

        self.add_feature_btn = QPushButton("")
        self.add_feature_btn.setStyleSheet(self.button_select_all)
        self.add_feature_btn.setIcon(icon(MDI6.filter_plus, color="black"))
        self.add_feature_btn.setToolTip("Add feature")
        self.add_feature_btn.setIconSize(QSize(20, 20))

        self.features_list = ListWidget(FeatureChoice, initial_features=['area', 'intensity_mean', ])

        self.del_feature_btn.clicked.connect(self.features_list.removeSel)
        self.add_feature_btn.clicked.connect(self.features_list.addItem)

        feature_layout.addWidget(self.feature_lbl, 90)
        feature_layout.addWidget(self.del_feature_btn, 5)
        feature_layout.addWidget(self.add_feature_btn, 5)
        layout.addLayout(feature_layout)
        layout.addWidget(self.features_list)

        self.feat_sep2 = QHSeperationLine()
        layout.addWidget(self.feat_sep2)

        contour_layout = QHBoxLayout()
        self.border_dist_lbl = QLabel('Contour measurements (from edge of mask):')
        self.border_dist_lbl.setToolTip(
            'Apply the intensity measurements defined above\nto a slice of each cell mask, defined as distance\nfrom the edge.')
        contour_layout.addWidget(self.border_dist_lbl, 90)

        self.del_contour_btn = QPushButton("")
        self.del_contour_btn.setStyleSheet(self.button_select_all)
        self.del_contour_btn.setIcon(icon(MDI6.trash_can, color="black"))
        self.del_contour_btn.setToolTip("Remove distance")
        self.del_contour_btn.setIconSize(QSize(20, 20))
        contour_layout.addWidget(self.del_contour_btn, 5)

        self.add_contour_btn = QPushButton("")
        self.add_contour_btn.setStyleSheet(self.button_select_all)
        self.add_contour_btn.setIcon(icon(MDI6.plus, color="black"))
        self.add_contour_btn.setToolTip("Add distance")
        self.add_contour_btn.setIconSize(QSize(20, 20))
        contour_layout.addWidget(self.add_contour_btn, 5)

        self.view_contour_btn = QPushButton("")
        self.view_contour_btn.setStyleSheet(self.button_select_all)
        self.view_contour_btn.setIcon(icon(MDI6.eye_plus_outline, color="black"))
        self.view_contour_btn.setToolTip("View contour")
        self.view_contour_btn.setIconSize(QSize(20, 20))
        contour_layout.addWidget(self.view_contour_btn, 5)

        layout.addLayout(contour_layout)

        self.contours_list = ListWidget(GeometryChoice, initial_features=[], dtype=int)
        layout.addWidget(self.contours_list)

        self.del_contour_btn.clicked.connect(self.contours_list.removeSel)
        self.add_contour_btn.clicked.connect(self.contours_list.addItem)
        self.view_contour_btn.clicked.connect(self.view_selected_contour)

        self.feat_sep3 = QHSeperationLine()
        layout.addWidget(self.feat_sep3)
        # self.radial_intensity_btn = QCheckBox('Measure radial intensity distribution')
        # layout.addWidget(self.radial_intensity_btn)
        # self.radial_intensity_btn.clicked.connect(self.enable_step_size)
        # self.channel_chechkboxes=[]
        # for channel in self.channel_names:
        #     channel_checkbox=QCheckBox(channel)
        #     self.channel_chechkboxes.append(channel_checkbox)
        #     layout.addWidget(channel_checkbox)
        #     channel_checkbox.setEnabled(False)
        # step_box=QHBoxLayout()
        # self.step_lbl=QLabel("Step size (in px)")
        # self.step_size=QLineEdit()
        # self.step_lbl.setEnabled(False)
        # self.step_size.setEnabled(False)
        # step_box.addWidget(self.step_lbl)
        # step_box.addWidget(self.step_size)
        # layout.addLayout(step_box)
        # self.feat_sep4 = QHSeperationLine()
        # layout.addWidget(self.feat_sep4)

        # Haralick features parameters
        self.activate_haralick_btn = QCheckBox('Measure Haralick texture features')
        self.activate_haralick_btn.toggled.connect(self.show_haralick_options)

        self.haralick_channel_choice = QComboBox()
        self.haralick_channel_choice.addItems(self.channel_names)
        self.haralick_channel_lbl = QLabel('Target channel: ')

        self.haralick_distance_le = QLineEdit("1")
        self.haralick_distance_lbl = QLabel('Distance: ')

        self.haralick_n_gray_levels_le = QLineEdit("256")
        self.haralick_n_gray_levels_lbl = QLabel('# gray levels: ')

        # Slider to set vmin & vmax
        self.haralick_scale_slider = QLabeledDoubleSlider()
        self.haralick_scale_slider.setSingleStep(0.05)
        self.haralick_scale_slider.setTickInterval(0.05)
        self.haralick_scale_slider.setSingleStep(1)
        self.haralick_scale_slider.setOrientation(1)
        self.haralick_scale_slider.setRange(0, 1)
        self.haralick_scale_slider.setValue(0.5)
        self.haralick_scale_lbl = QLabel('Scale: ')

        self.haralick_percentile_min_le = QLineEdit('0.01')
        self.haralick_percentile_max_le = QLineEdit('99.9')
        self.haralick_normalization_mode_btn = QPushButton()
        self.haralick_normalization_mode_btn.setStyleSheet(self.button_select_all)
        self.haralick_normalization_mode_btn.setIcon(icon(MDI6.percent_circle, color="black"))
        self.haralick_normalization_mode_btn.setIconSize(QSize(20, 20))
        self.haralick_normalization_mode_btn.setToolTip("Switch to absolute normalization values.")
        self.percentile_mode = True

        min_percentile_hbox = QHBoxLayout()
        min_percentile_hbox.addWidget(self.haralick_percentile_min_le, 90)
        min_percentile_hbox.addWidget(self.haralick_normalization_mode_btn, 10)
        min_percentile_hbox.setContentsMargins(0, 0, 0, 0)

        self.haralick_percentile_min_lbl = QLabel('Min percentile: ')
        self.haralick_percentile_max_lbl = QLabel('Max percentile: ')

        self.haralick_hist_btn = QPushButton()
        self.haralick_hist_btn.clicked.connect(self.control_haralick_intensity_histogram)
        self.haralick_hist_btn.setIcon(icon(MDI6.poll, color="k"))
        self.haralick_hist_btn.setStyleSheet(self.button_select_all)

        self.haralick_digit_btn = QPushButton()
        self.haralick_digit_btn.clicked.connect(self.control_haralick_digitalization)
        self.haralick_digit_btn.setIcon(icon(MDI6.image_check, color="k"))
        self.haralick_digit_btn.setStyleSheet(self.button_select_all)

        self.haralick_layout = QVBoxLayout()
        self.haralick_layout.setContentsMargins(20, 20, 20, 20)

        activate_layout = QHBoxLayout()
        activate_layout.addWidget(self.activate_haralick_btn, 80)
        activate_layout.addWidget(self.haralick_hist_btn, 10)
        activate_layout.addWidget(self.haralick_digit_btn, 10)
        self.haralick_layout.addLayout(activate_layout)

        channel_layout = QHBoxLayout()
        channel_layout.addWidget(self.haralick_channel_lbl, 40)
        channel_layout.addWidget(self.haralick_channel_choice, 60)
        self.haralick_layout.addLayout(channel_layout)

        distance_layout = QHBoxLayout()
        distance_layout.addWidget(self.haralick_distance_lbl, 40)
        distance_layout.addWidget(self.haralick_distance_le, 60)
        self.haralick_layout.addLayout(distance_layout)

        gl_layout = QHBoxLayout()
        gl_layout.addWidget(self.haralick_n_gray_levels_lbl, 40)
        gl_layout.addWidget(self.haralick_n_gray_levels_le, 60)
        self.haralick_layout.addLayout(gl_layout)

        slider_layout = QHBoxLayout()
        slider_layout.addWidget(self.haralick_scale_lbl, 40)
        slider_layout.addWidget(self.haralick_scale_slider, 60)
        self.haralick_layout.addLayout(slider_layout)

        slider_min_percentile_layout = QHBoxLayout()
        slider_min_percentile_layout.addWidget(self.haralick_percentile_min_lbl, 40)
        # slider_min_percentile_layout.addWidget(self.haralick_percentile_min_le,55)
        slider_min_percentile_layout.addLayout(min_percentile_hbox, 60)
        # slider_min_percentile_layout.addWidget(self.haralick_normalization_mode_btn, 5)
        self.haralick_layout.addLayout(slider_min_percentile_layout)

        slider_max_percentile_layout = QHBoxLayout()
        slider_max_percentile_layout.addWidget(self.haralick_percentile_max_lbl, 40)
        slider_max_percentile_layout.addWidget(self.haralick_percentile_max_le, 60)
        self.haralick_layout.addLayout(slider_max_percentile_layout)

        self.haralick_to_hide = [self.haralick_hist_btn, self.haralick_digit_btn, self.haralick_channel_lbl,
                                 self.haralick_channel_choice,
                                 self.haralick_distance_le, self.haralick_distance_lbl, self.haralick_n_gray_levels_le,
                                 self.haralick_n_gray_levels_lbl,
                                 self.haralick_scale_lbl, self.haralick_scale_slider, self.haralick_percentile_min_lbl,
                                 self.haralick_percentile_min_le,
                                 self.haralick_percentile_max_lbl, self.haralick_percentile_max_le,
                                 self.haralick_normalization_mode_btn]

        self.features_to_disable = [self.feature_lbl, self.del_feature_btn, self.add_feature_btn, self.features_list,
                                    self.activate_haralick_btn]

        self.activate_haralick_btn.setChecked(False)
        for f in self.haralick_to_hide:
            f.setEnabled(False)

        self.haralick_normalization_mode_btn.clicked.connect(self.switch_to_absolute_normalization_mode)
        layout.addLayout(self.haralick_layout)

    def switch_to_absolute_normalization_mode(self):

        if self.percentile_mode:
            self.percentile_mode = False
            self.haralick_normalization_mode_btn.setIcon(icon(MDI6.percent_circle_outline, color="black"))
            self.haralick_normalization_mode_btn.setIconSize(QSize(20, 20))
            self.haralick_normalization_mode_btn.setToolTip("Switch to percentile normalization values.")
            self.haralick_percentile_min_lbl.setText('Min value: ')
            self.haralick_percentile_max_lbl.setText('Max value: ')
            self.haralick_percentile_min_le.setText('0')
            self.haralick_percentile_max_le.setText('10000')

        else:
            self.percentile_mode = True
            self.haralick_normalization_mode_btn.setIcon(icon(MDI6.percent_circle, color="black"))
            self.haralick_normalization_mode_btn.setIconSize(QSize(20, 20))
            self.haralick_normalization_mode_btn.setToolTip("Switch to absolute normalization values.")
            self.haralick_percentile_min_lbl.setText('Min percentile: ')
            self.haralick_percentile_max_lbl.setText('Max percentile: ')
            self.haralick_percentile_min_le.setText('0.01')
            self.haralick_percentile_max_le.setText('99.99')

    def show_haralick_options(self):

        """
        Show the Haralick texture options.
        """

        if self.activate_haralick_btn.isChecked():
            for element in self.haralick_to_hide:
                element.setEnabled(True)
        else:
            for element in self.haralick_to_hide:
                element.setEnabled(False)

    def adjustScrollArea(self):

        """
        Auto-adjust scroll area to fill space
        (from https://stackoverflow.com/questions/66417576/make-qscrollarea-use-all-available-space-of-qmainwindow-height-axis)
        """

        step = 5
        while self.scroll_area.verticalScrollBar().isVisible() and self.height() < self.maximumHeight():
            self.resize(self.width(), self.height() + step)

    def write_instructions(self):

        """
        Write the selected options in a json file for later reading by the software.
        """

        print('Writing instructions...')
        measurement_options = {}
        background_correction = self.protocol_layout.protocols
        if not background_correction:
            background_correction = None
        measurement_options.update({'background_correction': background_correction})
        features = self.features_list.getItems()
        if not features:
            features = None
        measurement_options.update({'features': features})

        border_distances = self.contours_list.getItems()
        if not border_distances:
            border_distances = None
        measurement_options.update({'border_distances': border_distances})
        # radial_intensity = {}
        # radial_step = int(self.step_size.text())
        # radial_channels = []
        # for checkbox in self.channel_chechkboxes:
        #     if checkbox.isChecked():
        #         radial_channels.append(checkbox.text())
        # radial_intensity={'radial_step': radial_step, 'radial_channels': radial_channels}
        # if not self.radial_intensity_btn.isChecked():
        #     radial_intensity = None
        # measurement_options.update({'radial_intensity' : radial_intensity})

        self.extract_haralick_options()
        measurement_options.update({'haralick_options': self.haralick_options})

        intensity_measurement_radii = self.radii_list.getItems()
        if not intensity_measurement_radii:
            intensity_measurement_radii = None

        isotropic_operations = self.operations_list.getItems()
        if not isotropic_operations:
            isotropic_operations = None
            intensity_measurement_radii = None
        measurement_options.update({'intensity_measurement_radii': intensity_measurement_radii,
                                    'isotropic_operations': isotropic_operations})
        spot_detection = None
        if self.spot_check.isChecked():
            spot_detection = {'channel': self.spot_channel.currentText(), 'diameter': float(self.diameter_value.text().replace(',','.')),
                              'threshold': float(self.threshold_value.text().replace(',','.'))}
        measurement_options.update({'spot_detection': spot_detection})
        if self.clear_previous_btn.isChecked():
            self.clear_previous = True
        else:
            self.clear_previous = False
        measurement_options.update({'clear_previous': self.clear_previous})



        print('Measurement instructions: ', measurement_options)
        file_name = self.measure_instructions_path
        with open(file_name, 'w') as f:
            json.dump(measurement_options, f, indent=4)

        print('Done.')
        self.close()

    def extract_haralick_options(self):

        if self.activate_haralick_btn.isChecked():
            self.haralick_options = {"target_channel": self.haralick_channel_choice.currentIndex(),
                                     "scale_factor": float(self.haralick_scale_slider.value()),
                                     "n_intensity_bins": int(self.haralick_n_gray_levels_le.text()),
                                     "distance": int(self.haralick_distance_le.text()),
                                     }
            if self.percentile_mode:
                self.haralick_options.update({"percentiles": (
                    float(self.haralick_percentile_min_le.text()), float(self.haralick_percentile_max_le.text())),
                    "clip_values": None})
            else:
                self.haralick_options.update({"percentiles": None, "clip_values": (
                    float(self.haralick_percentile_min_le.text()), float(self.haralick_percentile_max_le.text()))})

        else:
            self.haralick_options = None

    def load_previous_measurement_instructions(self):

        """
        Read the measurmeent options from a previously written json file and format properly for the UI.
        """

        print('Reading instructions..')
        if os.path.exists(self.measure_instructions_path):
            with open(self.measure_instructions_path, 'r') as f:
                measurement_instructions = json.load(f)
                print(measurement_instructions)
                if 'background_correction' in measurement_instructions:
                    self.protocol_layout.protocols = measurement_instructions['background_correction']
                    if self.protocol_layout.protocols is None:
                        self.protocol_layout.protocols = []
                    if (self.protocol_layout.protocols is not None) and len(self.protocol_layout.protocols) > 0:
                        self.protocol_layout.protocol_list.clear()
                        for norm_params in self.protocol_layout.protocols:
                            normalisation_description = ""
                            for index, (key, value) in enumerate(norm_params.items()):
                                if index > 0:
                                    normalisation_description += ", "
                                normalisation_description += str(key) + " : " + str(value)
                            self.protocol_layout.protocol_list.addItem(normalisation_description)
                    else:
                        self.protocol_layout.protocol_list.clear()
                if 'features' in measurement_instructions:
                    features = measurement_instructions['features']
                    if (features is not None) and len(features) > 0:
                        self.features_list.list_widget.clear()
                        self.features_list.list_widget.addItems(features)
                    else:
                        self.features_list.list_widget.clear()

                if 'spot_detection' in measurement_instructions:
                    spot_detection = measurement_instructions['spot_detection']
                    if spot_detection is not None:
                        self.spot_check.setChecked(True)
                        if 'channel' in spot_detection:
                            idx = spot_detection['channel']
                            self.spot_channel.setCurrentText(idx)
                        self.diameter_value.setText(str(spot_detection['diameter']))
                        self.threshold_value.setText(str(spot_detection['threshold']))


                if 'border_distances' in measurement_instructions:
                    border_distances = measurement_instructions['border_distances']
                    if border_distances is not None:
                        if isinstance(border_distances, int):
                            distances = [border_distances]
                        elif isinstance(border_distances, list):
                            distances = []
                            for d in border_distances:
                                if isinstance(d, int) | isinstance(d, float):
                                    distances.append(str(int(d)))
                                elif isinstance(d, list):
                                    distances.append(str(int(d[0])) + '-' + str(int(d[1])))
                        self.contours_list.list_widget.clear()
                        self.contours_list.list_widget.addItems(distances)

                if 'haralick_options' in measurement_instructions:
                    haralick_options = measurement_instructions['haralick_options']
                    if haralick_options is None:
                        self.activate_haralick_btn.setChecked(False)
                        self.show_haralick_options()
                    else:
                        self.activate_haralick_btn.setChecked(True)
                        self.show_haralick_options()
                        if 'target_channel' in haralick_options:
                            idx = haralick_options['target_channel']
                            self.haralick_channel_choice.setCurrentIndex(idx)
                        if 'scale_factor' in haralick_options:
                            self.haralick_scale_slider.setValue(float(haralick_options['scale_factor']))
                        if ('percentiles' in haralick_options) and (haralick_options['percentiles'] is not None):
                            perc = list(haralick_options['percentiles'])
                            self.haralick_percentile_min_le.setText(str(perc[0]))
                            self.haralick_percentile_max_le.setText(str(perc[1]))
                        if ('clip_values' in haralick_options) and (haralick_options['clip_values'] is not None):
                            values = list(haralick_options['clip_values'])
                            self.haralick_percentile_min_le.setText(str(values[0]))
                            self.haralick_percentile_max_le.setText(str(values[1]))
                            self.percentile_mode = True
                            self.switch_to_absolute_normalization_mode()
                        if 'n_intensity_bins' in haralick_options:
                            self.haralick_n_gray_levels_le.setText(str(haralick_options['n_intensity_bins']))
                        if 'distance' in haralick_options:
                            self.haralick_distance_le.setText(str(haralick_options['distance']))

                if 'intensity_measurement_radii' in measurement_instructions:
                    intensity_measurement_radii = measurement_instructions['intensity_measurement_radii']
                    if intensity_measurement_radii is not None:
                        if isinstance(intensity_measurement_radii, int):
                            radii = [intensity_measurement_radii]
                        elif isinstance(intensity_measurement_radii, list):
                            radii = []
                            for r in intensity_measurement_radii:
                                if isinstance(r, int) | isinstance(r, float):
                                    radii.append(str(int(r)))
                                elif isinstance(r, list):
                                    radii.append(str(int(r[0])) + '-' + str(int(r[1])))
                        self.radii_list.list_widget.clear()
                        self.radii_list.list_widget.addItems(radii)
                    else:
                        self.radii_list.list_widget.clear()

                if 'isotropic_operations' in measurement_instructions:
                    isotropic_operations = measurement_instructions['isotropic_operations']
                    if (isotropic_operations is not None) and len(isotropic_operations) > 0:
                        self.operations_list.list_widget.clear()
                        self.operations_list.list_widget.addItems(isotropic_operations)
                    else:
                        self.operations_list.list_widget.clear()

                # if 'radial_intensity' in measurement_instructions:
                #     radial_intensity = measurement_instructions['radial_intensity']
                #     if radial_intensity is not None:
                #         self.radial_intensity_btn.setChecked(True)
                #         self.step_size.setText(str(radial_intensity['radial_step']))
                #         self.step_size.setEnabled(True)
                #         self.step_lbl.setEnabled(True)
                #         for checkbox in self.channel_chechkboxes:
                #             checkbox.setEnabled(True)
                #             if checkbox.text() in radial_intensity['radial_channels']:
                #                 checkbox.setChecked(True)

                if 'clear_previous' in measurement_instructions:
                    self.clear_previous = measurement_instructions['clear_previous']
                    self.clear_previous_btn.setChecked(self.clear_previous)

    def locate_image(self):

        """
        Load the first frame of the first movie found in the experiment folder as a sample.
        """

        movies = glob(self.parent_window.parent_window.pos + os.sep.join(['movie', f"{self.parent_window.movie_prefix}*.tif"]))

        if len(movies) == 0:
            msgBox = QMessageBox()
            msgBox.setIcon(QMessageBox.Warning)
            msgBox.setText("Please select a position containing a movie...")
            msgBox.setWindowTitle("Warning")
            msgBox.setStandardButtons(QMessageBox.Ok)
            returnValue = msgBox.exec()
            if returnValue == QMessageBox.Ok:
                self.current_stack = None
                return None
        else:
            self.current_stack = movies[0]
            self.stack_length = auto_load_number_of_frames(self.current_stack)
            
            if self.stack_length is None:
                stack = imread(self.current_stack)
                self.stack_length = len(stack)
                del stack
                gc.collect()
            
            self.mid_time = self.stack_length // 2
            indices = self.mid_time + np.arange(len(self.channel_names))
            self.test_frame = load_frames(list(indices.astype(int)),self.current_stack, normalize_input=False)


    def control_haralick_digitalization(self):

        """
        Load an image for the first experiment movie found.
        Apply the Haralick parameters and check the result of the digitization (normalization + binning of intensities).

        """

        self.locate_image()
        self.extract_haralick_options()
        if self.test_frame is not None:
            digitized_img = compute_haralick_features(self.test_frame, np.zeros(self.test_frame.shape[:2]),
                                                      channels=self.channel_names, return_digit_image_only=True,
                                                      **self.haralick_options
                                                      )

            self.fig, self.ax = plt.subplots()
            divider = make_axes_locatable(self.ax)
            cax = divider.append_axes('right', size='5%', pad=0.05)

            self.imshow_digit_window = FigureCanvas(self.fig, title="Haralick: control digitization")
            self.ax.clear()
            im = self.ax.imshow(digitized_img, cmap='gray')
            self.fig.colorbar(im, cax=cax, orientation='vertical')
            self.ax.set_xticks([])
            self.ax.set_yticks([])
            self.fig.set_facecolor('none')  # or 'None'
            self.fig.canvas.setStyleSheet("background-color: transparent;")
            self.imshow_digit_window.canvas.draw()
            self.imshow_digit_window.show()

    def control_haralick_intensity_histogram(self):

        """
        Load an image for the first experiment movie found.
        Apply the Haralick normalization parameters and check the normalized intensity histogram.

        """

        self.locate_image()
        self.extract_haralick_options()
        if self.test_frame is not None:
            norm_img = compute_haralick_features(self.test_frame, np.zeros(self.test_frame.shape[:2]),
                                                 channels=self.channel_names, return_norm_image_only=True,
                                                 **self.haralick_options
                                                 )
            self.fig, self.ax = plt.subplots(1, 1, figsize=(4, 3))
            self.hist_window = FigureCanvas(self.fig, title="Haralick: control digitized histogram")
            self.ax.clear()
            flat = norm_img.flatten()
            self.ax.hist(flat[flat==flat], bins=self.haralick_options['n_intensity_bins'])
            self.ax.set_xlabel('gray level value')
            self.ax.set_ylabel('#')
            plt.tight_layout()
            self.fig.set_facecolor('none')  # or 'None'
            self.fig.canvas.setStyleSheet("background-color: transparent;")
            self.hist_window.canvas.draw()
            self.hist_window.show()

    def view_selected_contour(self):

        """
        Show the ROI for the selected contour measurement on experimental data.

        """
        self.locate_image()

        if self.current_stack is not None:

            self.viewer = CellEdgeVisualizer(cell_type=self.mode,
                                             stack_path=self.current_stack,
                                             parent_list_widget=self.contours_list.list_widget,
                                             n_channels=len(self.channel_names),
                                             target_channel=0,
                                             window_title='Set an edge measurement',
                                             channel_cb=True,
                                             channel_names = self.channel_names,
                                             PxToUm = 1,
                                             )
            self.viewer.show()

    def locate_mask(self):

        """
        Load the first mask of the detected movie.
        """

        labels_path = str(Path(self.current_stack).parent.parent) + os.sep+f'labels_{self.mode}'+os.sep
        masks = natsorted(glob(labels_path + '*.tif'))
        if len(masks) == 0:
            print('no mask found')
            self.test_mask = None
        else:
            self.test_mask = imread(masks[self.mid_time])

    def switch_channel_contour(self, value):

        """
        Adjust intensity values when changing channels in the contour visualizer.

        """

        self.im_contour.set_array(self.test_frame[:, :, value])
        self.im_contour.set_clim(vmin=np.percentile(self.test_frame[:, :, value].flatten(), 1),
                                 vmax=np.percentile(self.test_frame[:, :, value].flatten(), 99.99))
        self.contrast_slider_contour.setRange(np.amin(self.test_frame[:, :, value]),
                                              np.amax(self.test_frame[:, :, value]))
        self.contrast_slider_contour.setValue([np.percentile(self.test_frame[:, :, value].flatten(), 1),
                                               np.percentile(self.test_frame[:, :, value].flatten(), 99.99)])
        self.fig_contour.canvas.draw_idle()

    def contrast_im_contour(self, value):
        vmin = value[0]
        vmax = value[1]
        self.im_contour.set_clim(vmin=vmin, vmax=vmax)
        self.fig_contour.canvas.draw_idle()

    def make_contour_transparent(self, value):

        self.im_mask.set_alpha(value)
        self.fig_contour.canvas.draw_idle()

    def remove_item_from_list(self):
        current_item = self.normalisation_list.currentRow()
        if current_item > -1:
            del self.background_correction[current_item]
            self.normalisation_list.takeItem(current_item)

    def check_the_information(self):
        
        if self.tabs.currentIndex() == 0:
            if self.background_correction  is None:
                self.background_correction  = []
            for index, normalisation_opt in enumerate(self.background_correction ):
                if self.tab1_channel_dropdown.currentText() in normalisation_opt['target channel']:
                    result = self.channel_already_in_list()
                    if result != QMessageBox.Yes:
                        return False
                    else:
                        self.background_correction .remove(normalisation_opt)
                        self.normalisation_list.takeItem(index)
                        return True

    def display_message_box(self, missing_info):
        QMessageBox.about(self, "Message Box Title", "Please " + missing_info + " for background correction")

    def channel_already_in_list(self):
        response = QMessageBox.question(self, "Message Box Title",
                                        "The background correction parameters for this channel already exist, "
                                        "continuing will erase the previous configurations. Are you sure you want to "
                                        "proceed?",
                                        QMessageBox.No | QMessageBox.Yes, QMessageBox.No)
        return response

    def fun(self, x, y):
        return x ** 2 + y

    def view_normalisation_contour(self):

        """
        Show the ROI for the selected contour measurement on experimental data.

        """

        self.locate_image()

        if self.current_stack is not None:

            self.viewer = CellEdgeVisualizer(cell_type=self.mode,
                                             stack_path=self.current_stack,
                                             parent_le = self.tab1_txt_distance,
                                             n_channels=len(self.channel_names),
                                             target_channel=self.tab1_channel_dropdown.currentIndex(),
                                             edge_range = (0,30),
                                             initial_edge= self.tab1_txt_distance.get_threshold(),
                                             invert=True,
                                             window_title='Set an edge distance to estimate local intensity',
                                             channel_cb=False,
                                             PxToUm = 1,
                                             )
            self.viewer.show()


    def populate_spot_detection(self):

        layout = QGridLayout(self.spot_detection_frame)
        self.spot_detection_lbl = QLabel("SPOT DETECTION")
        self.spot_detection_lbl.setStyleSheet("""font-weight: bold;padding: 0px;""")
        layout.addWidget(self.spot_detection_lbl, 0, 0, 1, 2, alignment=Qt.AlignCenter)
        self.spot_check= QCheckBox('Perform spot detection')
        self.spot_check.toggled.connect(self.enable_spot_detection)
        layout.addWidget(self.spot_check, 1, 0)
        self.spot_channel_lbl = QLabel("Choose channel for spot detection: ")
        self.spot_channel = QComboBox()
        self.spot_channel.addItems(self.channel_names)
        layout.addWidget(self.spot_channel_lbl, 2, 0)
        layout.addWidget(self.spot_channel, 2, 1)
        self.diameter_lbl = QLabel('Spot diameter: ')
        self.diameter_value = QLineEdit()
        self.diameter_value.setValidator(self.onlyFloat)
        self.diameter_value.setText('7')
        self.diameter_value.textChanged.connect(self.enable_spot_preview)

        layout.addWidget(self.diameter_lbl, 3, 0)
        layout.addWidget(self.diameter_value, 3, 1)
        self.threshold_lbl = QLabel('Spot threshold: ')
        self.threshold_value = QLineEdit()
        self.threshold_value.setValidator(self.onlyFloat)
        self.threshold_value.setText('0')
        self.threshold_value.textChanged.connect(self.enable_spot_preview)

        layout.addWidget(self.threshold_lbl, 4, 0)
        layout.addWidget(self.threshold_value, 4, 1)
        self.preview_spot = QPushButton('Preview')
        self.preview_spot.clicked.connect(self.spot_preview)
        self.preview_spot.setStyleSheet(self.button_style_sheet_2)
        layout.addWidget(self.preview_spot, 5, 0, 1, 2)
        self.spot_channel.setEnabled(False)
        self.spot_channel_lbl.setEnabled(False)
        self.diameter_value.setEnabled(False)
        self.diameter_lbl.setEnabled(False)
        self.threshold_value.setEnabled(False)
        self.threshold_lbl.setEnabled(False)
        self.preview_spot.setEnabled(False)


    def enable_spot_preview(self):

        diam = self.diameter_value.text().replace(',','').replace('.','')
        thresh = self.threshold_value.text().replace(',','').replace('.','')
        if diam.isnumeric() and thresh.isnumeric():
            self.preview_spot.setEnabled(True)
        else:
            self.preview_spot.setEnabled(False)

    def spot_preview(self):
        self.locate_image()
        if self.test_frame is not None:
            self.locate_mask()
            if self.test_mask is not None:
                self.spot_visual = ThresholdSpot(current_channel=self.spot_channel.currentIndex(), img=self.test_frame,
                                                 mask=self.test_mask, parent_window=self)

    def enable_spot_detection(self):
        if self.spot_check.isChecked():
            self.spot_channel.setEnabled(True)
            self.spot_channel_lbl.setEnabled(True)
            self.diameter_value.setEnabled(True)
            self.diameter_lbl.setEnabled(True)
            self.threshold_value.setEnabled(True)
            self.threshold_lbl.setEnabled(True)
            self.preview_spot.setEnabled(True)

        else:
            self.spot_channel.setEnabled(False)
            self.spot_channel_lbl.setEnabled(False)
            self.diameter_value.setEnabled(False)
            self.diameter_lbl.setEnabled(False)
            self.threshold_value.setEnabled(False)
            self.threshold_lbl.setEnabled(False)
            self.preview_spot.setEnabled(False)



import argparse
import os
from celldetective.relative_measurements import measure_pair_signals_at_position, extract_neighborhoods_from_pickles
from celldetective.utils import ConfigSectionMap, extract_experiment_channels

from pathlib import Path, PurePath

import pandas as pd

from art import tprint


tprint("Measure pairs")

parser = argparse.ArgumentParser(description="Measure features and intensities in a multichannel timeseries.",
								 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-p', "--position", required=True, help="Path to the position")

args = parser.parse_args()
process_arguments = vars(args)
pos = str(process_arguments['position'])

instruction_file = os.sep.join(['configs', "neighborhood_instructions.json"])

# Locate experiment config
parent1 = Path(pos).parent
expfolder = parent1.parent
config = PurePath(expfolder, Path("config.ini"))
assert os.path.exists(config), 'The configuration file for the experiment could not be located. Abort.'
print("Configuration file: ", config)

# from exp config fetch spatial calib, channel names
movie_prefix = ConfigSectionMap(config, "MovieSettings")["movie_prefix"]
spatial_calibration = float(ConfigSectionMap(config, "MovieSettings")["pxtoum"])
time_calibration = float(ConfigSectionMap(config, "MovieSettings")["frametomin"])
len_movie = float(ConfigSectionMap(config, "MovieSettings")["len_movie"])
channel_names, channel_inneigh_protocoles = extract_experiment_channels(config)
nbr_channels = len(channel_names)

# from tracking instructions, fetch btrack config, features, haralick, clean_traj, idea: fetch custom timeline?
instr_path = PurePath(expfolder, Path(f"{instruction_file}"))
previous_pair_table_path = pos + os.sep.join(['output', 'tables', 'trajectories_pairs.csv'])

# if os.path.exists(instr_path):
# 	print(f"Neighborhood instructions has been successfully located.")
# 	with open(instr_path, 'r') as f:
# 		instructions = json.load(f)
# 		print("Reading the following instructions: ", instructions)

# 	if 'distance' in instructions:
# 		distance = instructions['distance'][0]
# 	else:
# 		distance = None
# else:
# 	print('No measurement instructions found')
# 	os.abort()

previous_neighborhoods = []
associated_reference_population = []

# if distance is None:
# 	print('No measurement could be performed. Check your inputs.')
# 	print('Done.')
# 	os.abort()
# 	#distance = 0
# else:
neighborhoods_to_measure = extract_neighborhoods_from_pickles(pos)
all_df_pairs = []
if os.path.exists(previous_pair_table_path):
	df_0 = pd.read_csv(previous_pair_table_path)
	previous_neighborhoods = [c.replace('status_','') for c in list(df_0.columns) if c.startswith('status_neighborhood')]
	for n in previous_neighborhoods:
		associated_reference_population.append(df_0.loc[~df_0['status_'+n].isnull(),'reference_population'].values[0])
	print(f'{previous_neighborhoods=} {associated_reference_population=}')
	all_df_pairs.append(df_0)
for k,neigh_protocol in enumerate(neighborhoods_to_measure):
	if neigh_protocol['description'] not in previous_neighborhoods:
		df_pairs = measure_pair_signals_at_position(pos, neigh_protocol)
		print(f'{df_pairs=}')
		if 'REFERENCE_ID' in list(df_pairs.columns):
			all_df_pairs.append(df_pairs)
	elif neigh_protocol['description'] in previous_neighborhoods and neigh_protocol['reference'] != associated_reference_population[previous_neighborhoods.index(neigh_protocol['description'])]:
		df_pairs = measure_pair_signals_at_position(pos, neigh_protocol)
		if 'REFERENCE_ID' in list(df_pairs.columns):
			all_df_pairs.append(df_pairs)			

print(f'{len(all_df_pairs)} neighborhood measurements sets were computed...')

if len(all_df_pairs)>1:
	print('Merging...')
	df_pairs = all_df_pairs[0]
	for i in range(1,len(all_df_pairs)):
		cols = [c1 for c1,c2 in zip(list(df_pairs.columns), list(all_df_pairs[i].columns)) if c1==c2]
		df_pairs = pd.merge(df_pairs.round(decimals=6), all_df_pairs[i].round(decimals=6), how="outer", on=cols)
elif len(all_df_pairs)==1:
	df_pairs = all_df_pairs[0]

print('Writing table...')
df_pairs = df_pairs.sort_values(by=['reference_population', 'neighbor_population', 'REFERENCE_ID', 'NEIGHBOR_ID', 'FRAME'])
df_pairs.to_csv(previous_pair_table_path, index=False)
print('Done.')




import pandas as pd
import numpy as np
from tqdm import tqdm
from sklearn.preprocessing import StandardScaler

from btrack.io.utils import localizations_to_objects
from btrack import BayesianTracker

from celldetective.measure import measure_features
from celldetective.utils import rename_intensity_column, velocity_per_track
from celldetective.io import view_on_napari_btrack, interpret_tracking_configuration

import os
import subprocess

abs_path = os.sep.join([os.path.split(os.path.dirname(os.path.realpath(__file__)))[0],'celldetective'])

def track(labels, configuration=None, stack=None, spatial_calibration=1, features=None, channel_names=None,
		  haralick_options=None, return_napari_data=False, view_on_napari=False, mask_timepoints=None, mask_channels=None, volume=(2048,2048),
		  optimizer_options = {'tm_lim': int(12e4)}, track_kwargs={'step_size': 100}, objects=None,
		  clean_trajectories_kwargs=None, column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'},
		  ):

	"""

	Perform cell tracking on segmented labels using the bTrack library.

	Parameters
	----------
	labels : ndarray
		The segmented labels representing cell objects.
	configuration : Configuration or None
		The bTrack configuration object. If None, a default configuration is used.
	stack : ndarray or None, optional
		The image stack corresponding to the labels. Default is None.
	spatial_calibration : float, optional
		The spatial calibration factor to convert pixel coordinates to physical units. Default is 1.
	features : list or None, optional
		The list of features to extract from the objects. If None, no additional features are extracted. Default is None.
	channel_names : list or None, optional
		The list of channel names corresponding to the image stack. Used for renaming intensity columns in the output DataFrame.
		Default is None.
	haralick_options : dict or None, optional
		The options for Haralick feature extraction. If None, no Haralick features are extracted. Default is None.
	return_napari_data : bool, optional
		Whether to return the napari data dictionary along with the DataFrame. Default is False.
	view_on_napari : bool, optional
		Whether to view the tracking results on napari. Default is False.
	optimizer_options : dict, optional
		The options for the optimizer. Default is {'tm_lim': int(12e4)}.
	track_kwargs : dict, optional
		Additional keyword arguments for the bTrack tracker. Default is {'step_size': 100}.
	clean_trajectories_kwargs : dict or None, optional
		Keyword arguments for the clean_trajectories function to post-process the tracking trajectories. If None, no post-processing is performed.
		Default is None.
	column_labels : dict, optional
		The column labels to use in the output DataFrame. Default is {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	DataFrame or tuple
		If return_napari_data is False, returns the DataFrame containing the tracking results. If return_napari_data is True, returns a tuple
		containing the DataFrame and the napari data dictionary.

	Notes
	-----
	This function performs cell tracking on the segmented labels using the bTrack library. It extracts features from the objects, normalizes
	the features, tracks the objects, and generates a DataFrame with the tracking results. The DataFrame can be post-processed using the
	clean_trajectories function. If specified, the tracking results can be visualized on napari.

	Examples
	--------
	>>> labels = np.array([[1, 1, 2, 2, 0, 0],
						   [1, 1, 1, 2, 2, 0],
						   [0, 0, 1, 2, 0, 0]])
	>>> configuration = cell_config()
	>>> stack = np.random.rand(3, 6)
	>>> df = track(labels, configuration, stack=stack, spatial_calibration=0.5)
	>>> df.head()

	   TRACK_ID  FRAME  POSITION_Y  POSITION_X
	0         0      0         0.0         0.0
	1         0      1         0.0         0.0
	2         0      2         0.0         0.0
	3         1      0         0.5         0.5
	4         1      1         0.5         0.5

	"""

	configuration = interpret_tracking_configuration(configuration)

	if objects is None:
		objects = extract_objects_and_features(labels, stack, features, 
										   channel_names=channel_names,
										   haralick_options=haralick_options,
										   mask_timepoints=mask_timepoints,
										   mask_channels=mask_channels,
										   )

	columns = list(objects.columns)
	to_remove = ['x','y','class_id','t']
	for tr in to_remove:
		try:
			columns.remove(tr)
		except:
			print(f'column {tr} could not be found...')

	scaler = StandardScaler()
	if columns:
		x = objects[columns].values
		x_scaled = scaler.fit_transform(x)
		df_temp = pd.DataFrame(x_scaled, columns=columns, index = objects.index)
		objects[columns] = df_temp
	else:
		print('Warning: no features were passed to bTrack...')

	# 2) track the objects
	new_btrack_objects = localizations_to_objects(objects)

	with BayesianTracker() as tracker:

		tracker.configure(configuration)

		if columns:
			tracking_updates = ["motion","visual"]
			#tracker.tracking_updates = ["motion","visual"]
			tracker.features = columns
		else:
			tracking_updates = ["motion"]
		
		tracker.append(new_btrack_objects)
		tracker.volume = ((0,volume[0]), (0,volume[1]), (-1e5, 1e5)) #(-1e5, 1e5)
		#print(tracker.volume)
		tracker.track(tracking_updates=tracking_updates, **track_kwargs)
		tracker.optimize(options=optimizer_options)

		data, properties, graph = tracker.to_napari() #ndim=2

	# do the table post processing and napari options
	if data.shape[1]==4:
		df = pd.DataFrame(data, columns=[column_labels['track'],column_labels['time'],column_labels['y'],column_labels['x']])
	elif data.shape[1]==5:
		df = pd.DataFrame(data, columns=[column_labels['track'],column_labels['time'],"z",column_labels['y'],column_labels['x']])
		df = df.drop(columns=['z'])	
	df[column_labels['x']+'_um'] = df[column_labels['x']]*spatial_calibration
	df[column_labels['y']+'_um'] = df[column_labels['y']]*spatial_calibration

	df = df.merge(pd.DataFrame(properties),left_index=True, right_index=True)
	if columns:
		x = df[columns].values
		x_scaled = scaler.inverse_transform(x)
		df_temp = pd.DataFrame(x_scaled, columns=columns, index = df.index)
		df[columns] = df_temp

	# set dummy features to NaN
	df.loc[df['dummy'],['class_id']+columns] = np.nan 
	df = df.sort_values(by=[column_labels['track'],column_labels['time']])
	df = velocity_per_track(df, window_size=3, mode='bi')

	if channel_names is not None:
		df = rename_intensity_column(df, channel_names)

	df = write_first_detection_class(df, column_labels=column_labels)

	if clean_trajectories_kwargs is not None:
		df = clean_trajectories(df.copy(),**clean_trajectories_kwargs)

	df['ID'] = np.arange(len(df)).astype(int)

	if view_on_napari:
		view_on_napari_btrack(data,properties,graph,stack=stack,labels=labels,relabel=True)

	if return_napari_data:
		napari_data = {"data": data, "properties": properties, "graph": graph}
		return df, napari_data
	else:
		return df

def extract_objects_and_features(labels, stack, features, channel_names=None, haralick_options=None, mask_timepoints=None, mask_channels=None):

	"""

	Extract objects and features from segmented labels and image stack.

	Parameters
	----------
	labels : ndarray
		The segmented labels representing cell objects.
	stack : ndarray
		The image stack corresponding to the labels.
	features : list or None
		The list of features to extract from the objects. If None, no additional features are extracted.
	channel_names : list or None, optional
		The list of channel names corresponding to the image stack. Used for extracting Haralick features. Default is None.
	haralick_options : dict or None, optional
		The options for Haralick feature extraction. If None, no Haralick features are extracted. Default is None.
	mask_timepoints : list of None, optionak
		Frames to hide during tracking. 
	Returns
	-------
	DataFrame
		The DataFrame containing the extracted object features.

	Notes
	-----
	This function extracts objects and features from the segmented labels and image stack. It computes the specified features for each
	labeled object and returns a DataFrame containing the object features. Additional features such as centroid coordinates can also
	be extracted. If Haralick features are enabled, they are computed based on the image stack using the specified options.

	Examples
	--------
	>>> labels = np.array([[1, 1, 2, 2, 0, 0],
						   [1, 1, 1, 2, 2, 0],
						   [0, 0, 1, 2, 0, 0]])
	>>> stack = np.random.rand(3, 6, 3)
	>>> features = ['area', 'mean_intensity']
	>>> df = extract_objects_and_features(labels, stack, features)
	
	"""

	if features is None:
		features = []

	if stack is None:
		haralick_options = None

	if mask_timepoints is not None:
		for f in mask_timepoints:
			labels[f] = 0.

	nbr_frames = len(labels)
	timestep_dataframes = []

	for t in tqdm(range(nbr_frames),desc='frame'):

		if stack is not None:
			img = stack[t]
		else:
			img = None

		if (haralick_options is not None) and (t==0) and (stack is not None):
			if not 'percentiles' in haralick_options:
				haralick_options.update({'percentiles': (0.01,99.99)})
			if not 'target_channel' in haralick_options:
				haralick_options.update({'target_channel': 0})
			haralick_percentiles = haralick_options['percentiles']
			haralick_channel_index = haralick_options['target_channel']
			min_value = np.nanpercentile(img[:,:,haralick_channel_index].flatten(), haralick_percentiles[0])
			max_value = np.nanpercentile(img[:,:,haralick_channel_index].flatten(), haralick_percentiles[1])			
			haralick_options.update({'clip_values': (min_value, max_value)})

		df_props = measure_features(img, labels[t], features = features+['centroid'], border_dist=None, 
										channels=channel_names, haralick_options=haralick_options, verbose=False)
		df_props.rename(columns={'centroid-1': 'x', 'centroid-0': 'y'},inplace=True)
		df_props['t'] = int(t)
		timestep_dataframes.append(df_props)

	df = pd.concat(timestep_dataframes)	
	df.reset_index(inplace=True, drop=True)

	if mask_channels is not None:
		cols_to_drop = []
		for mc in mask_channels:
			columns = df.columns
			col_contains = [mc in c for c in columns]
			to_remove = np.array(columns)[np.array(col_contains)]
			cols_to_drop.extend(to_remove)
		if len(cols_to_drop)>0:
			df = df.drop(cols_to_drop, axis=1)

	return df


def clean_trajectories(trajectories,remove_not_in_first=False,remove_not_in_last=False,
					   minimum_tracklength=0, interpolate_position_gaps=False,
					   extrapolate_tracks_post=False,
					   extrapolate_tracks_pre=False,
					   interpolate_na=False,
					   column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):

	"""
	Clean trajectories by applying various cleaning operations.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	remove_not_in_first : bool, optional
		Flag indicating whether to remove tracks not present in the first frame.
		Defaults to True.
	remove_not_in_last : bool, optional
		Flag indicating whether to remove tracks not present in the last frame.
		Defaults to True.
	minimum_tracklength : int, optional
		The minimum length of a track to be retained.
		Defaults to 0.
	interpolate_position_gaps : bool, optional
		Flag indicating whether to interpolate position gaps in tracks.
		Defaults to True.
	extrapolate_tracks_post : bool, optional
		Flag indicating whether to extrapolate tracks after the last known position.
		Defaults to True.
	extrapolate_tracks_pre : bool, optional
		Flag indicating whether to extrapolate tracks before the first known position.
		Defaults to False.
	interpolate_na : bool, optional
		Flag indicating whether to interpolate missing values in tracks.
		Defaults to False.
	column_labels : dict, optional
		Dictionary specifying the column labels used in the input DataFrame.
		The keys represent the following column labels:
		- 'track': The column label for the track ID.
		- 'time': The column label for the timestamp.
		- 'x': The column label for the x-coordinate.
		- 'y': The column label for the y-coordinate.
		Defaults to {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	pandas.DataFrame
		The cleaned DataFrame with trajectories.

	Notes
	-----
	This function applies various cleaning operations to the input DataFrame containing trajectory data.
	The cleaning operations include:
	- Filtering tracks based on their endpoints.
	- Filtering tracks based on their length.
	- Interpolating position gaps in tracks.
	- Extrapolating tracks after the last known position.
	- Extrapolating tracks before the first known position.
	- Interpolating missing values in tracks.

	The input DataFrame is expected to have the following columns:
	- track: The unique ID of each track.
	- time: The timestamp of each data point.
	- x: The x-coordinate of each data point.
	- y: The y-coordinate of each data point.

	Examples
	--------
	>>> cleaned_data = clean_trajectories(trajectories, remove_not_in_first=True, remove_not_in_last=True,
	...                                   minimum_tracklength=10, interpolate_position_gaps=True,
	...                                   extrapolate_tracks_post=True, extrapolate_tracks_pre=False,
	...                                   interpolate_na=True, column_labels={'track': "ID", 'time': 'TIME', 'x': 'X', 'y': 'Y'})
	>>> print(cleaned_data.head())

	"""

	trajectories.reset_index
	trajectories.sort_values(by=[column_labels['track'],column_labels['time']],inplace=True)
	
	if minimum_tracklength>0:
		trajectories = filter_by_tracklength(trajectories.copy(), minimum_tracklength, track_label=column_labels['track'])
	
	if np.any([remove_not_in_first, remove_not_in_last]):
		trajectories = filter_by_endpoints(trajectories.copy(), remove_not_in_first=remove_not_in_first,
		remove_not_in_last=remove_not_in_last, column_labels=column_labels)

	if np.any([extrapolate_tracks_post, extrapolate_tracks_pre]):
		trajectories = extrapolate_tracks(trajectories.copy(), post=extrapolate_tracks_post,
		pre=extrapolate_tracks_pre, column_labels=column_labels)

	if interpolate_position_gaps:
		trajectories = interpolate_time_gaps(trajectories.copy(), column_labels=column_labels)    

	if interpolate_na:
		trajectories = interpolate_nan_properties(trajectories.copy(), track_label=column_labels['track'])
	
	trajectories = trajectories.sort_values(by=[column_labels['track'],column_labels['time']])
	trajectories.reset_index(inplace=True, drop=True)

	if 'class_firstdetection' in list(trajectories.columns):
		for tid, track_group in trajectories.groupby(column_labels['track']):
			indices = track_group.index

			class_values = np.array(track_group['class_firstdetection'].unique())
			class_values = class_values[class_values==class_values]
			t_values = np.array(track_group['t_firstdetection'].unique())
			t_values = t_values[t_values==t_values]
			if len(class_values)==0:
				class_values = 2
				t_values = -1
			else:
				class_values = class_values[0]
				t_values = t_values[0]

			trajectories.loc[indices, 'class_firstdetection'] = class_values
			trajectories.loc[indices, 't_firstdetection'] = t_values

	return trajectories

def interpolate_per_track(group_df):

	"""
	Interpolate missing values within a track.

	Parameters
	----------
	group_df : pandas.DataFrame
		The input DataFrame containing data for a single track.

	Returns
	-------
	pandas.DataFrame
		The interpolated DataFrame with missing values filled.

	Notes
	-----
	This function performs linear interpolation to fill missing values within a track.
	Missing values are interpolated based on the neighboring data points in the track.

	"""

	interpolated_group = group_df.interpolate(method='linear',limit_direction="both")

	return interpolated_group

def interpolate_nan_properties(trajectories, track_label="TRACK_ID"):

	"""
	Interpolate missing values within tracks in the input DataFrame.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	track_label : str, optional
		The column label for the track ID.
		Defaults to "TRACK_ID".

	Returns
	-------
	pandas.DataFrame
		The DataFrame with missing values interpolated within tracks.

	Notes
	-----
	This function groups the input DataFrame by track ID and applies `interpolate_per_track` function
	to interpolate missing values within each track.
	Missing values are interpolated based on the neighboring data points in each track.

	The input DataFrame is expected to have a column with the specified `track_label` containing the track IDs.

	Examples
	--------
	>>> interpolated_data = interpolate_nan_properties(trajectories, track_label="ID")
	>>> print(interpolated_data.head())

	"""

	trajectories = trajectories.groupby(track_label, group_keys=False).apply(interpolate_per_track)

	return trajectories


def filter_by_endpoints(trajectories, remove_not_in_first=True, remove_not_in_last=False,
						column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):

	"""
	Filter trajectories based on their endpoints.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	remove_not_in_first : bool, optional
		Flag indicating whether to remove tracks not present in the first frame.
		Defaults to True.
	remove_not_in_last : bool, optional
		Flag indicating whether to remove tracks not present in the last frame.
		Defaults to False.
	column_labels : dict, optional
		Dictionary specifying the column labels used in the input DataFrame.
		The keys represent the following column labels:
		- 'track': The column label for the track ID.
		- 'time': The column label for the timestamp.
		- 'x': The column label for the x-coordinate.
		- 'y': The column label for the y-coordinate.
		Defaults to {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	pandas.DataFrame
		The filtered DataFrame with trajectories based on their endpoints.

	Notes
	-----
	This function filters the input DataFrame based on the endpoints of the trajectories.
	The filtering can be performed in three modes:
	- remove_not_in_first=True and remove_not_in_last=False: Remove tracks that are not present in the first frame.
	- remove_not_in_first=False and remove_not_in_last=True: Remove tracks that are not present in the last frame.
	- remove_not_in_first=True and remove_not_in_last=True: Remove tracks that are not present in both the first and last frames.

	The input DataFrame is expected to have the following columns:
	- track: The unique ID of each track.
	- time: The timestamp of each data point.
	- x: The x-coordinate of each data point.
	- y: The y-coordinate of each data point.

	Examples
	--------
	>>> filtered_data = filter_by_endpoints(trajectories, remove_not_in_first=True, remove_not_in_last=False, column_labels={'track': "ID", 'time': 'TIME', 'x': 'X', 'y': 'Y'})
	>>> print(filtered_data.head())

	"""

	if (remove_not_in_first)*(not remove_not_in_last):
		# filter tracks not in first frame
		leftover_tracks = trajectories.groupby(column_labels['track']).min().index[trajectories.groupby(column_labels['track']).min()[column_labels['time']]==np.amin(trajectories[column_labels['time']])]
		trajectories = trajectories.loc[trajectories[column_labels['track']].isin(leftover_tracks)]
		
	elif (remove_not_in_last)*(not remove_not_in_first):
		# filter tracks not in last frame
		leftover_tracks = trajectories.groupby(column_labels['track']).max().index[trajectories.groupby(column_labels['track']).max()[column_labels['time']]==np.amax(trajectories[column_labels['time']])]
		trajectories = trajectories.loc[trajectories[column_labels['track']].isin(leftover_tracks)]
	  
	elif remove_not_in_first*remove_not_in_last:
		# filter tracks both not in first and last frame
		leftover_tracks = trajectories.groupby(column_labels['track']).max().index[(trajectories.groupby(column_labels['track']).max()[column_labels['time']]==np.amax(trajectories[column_labels['time']]))*(trajectories.groupby(column_labels['track']).min()[column_labels['time']]==np.amin(trajectories[column_labels['time']]))]
		trajectories = trajectories.loc[trajectories[column_labels['track']].isin(leftover_tracks)]

	trajectories = trajectories.sort_values(by=[column_labels['track'],column_labels['time']])

	return trajectories

def filter_by_tracklength(trajectories, minimum_tracklength, track_label="TRACK_ID"):
	
	"""
	Filter trajectories based on the minimum track length.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	minimum_tracklength : int
		The minimum length required for a track to be included.
	track_label : str, optional
		The column name in the DataFrame that represents the track ID.
		Defaults to "TRACK_ID".

	Returns
	-------
	pandas.DataFrame
		The filtered DataFrame with trajectories that meet the minimum track length.

	Notes
	-----
	This function removes any tracks from the input DataFrame that have a length
	(number of data points) less than the specified minimum track length.

	Examples
	--------
	>>> filtered_data = filter_by_tracklength(trajectories, 10, track_label="TrackID")
	>>> print(filtered_data.head())

	"""
	
	if minimum_tracklength>0:
		
		leftover_tracks = trajectories.groupby(track_label, group_keys=False).size().index[trajectories.groupby(track_label, group_keys=False).size() > minimum_tracklength]
		trajectories = trajectories.loc[trajectories[track_label].isin(leftover_tracks)]
	
	trajectories = trajectories.reset_index(drop=True)
	
	return trajectories
	

def interpolate_time_gaps(trajectories, column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):

	"""
	Interpolate time gaps in trajectories.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	column_labels : dict, optional
		Dictionary specifying the column labels used in the input DataFrame.
		The keys represent the following column labels:
		- 'track': The column label for the track ID.
		- 'time': The column label for the timestamp.
		- 'x': The column label for the x-coordinate.
		- 'y': The column label for the y-coordinate.
		Defaults to {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	pandas.DataFrame
		The interpolated DataFrame with reduced time gaps in trajectories.

	Notes
	-----
	This function performs interpolation on the input trajectories to reduce time gaps between data points.
	It uses linear interpolation to fill missing values for the specified x and y coordinate attributes.

	The input DataFrame is expected to have the following columns:
	- track: The unique ID of each track.
	- time: The timestamp of each data point (in seconds).
	- x: The x-coordinate of each data point.
	- y: The y-coordinate of each data point.

	Examples
	--------
	>>> interpolated_data = interpolate_time_gaps(trajectories, column_labels={'track': "ID", 'time': 'TIME', 'x': 'X', 'y': 'Y'})
	>>> print(interpolated_data.head())

	"""

	trajectories[column_labels['time']] = pd.to_datetime(trajectories[column_labels['time']], unit='s')
	trajectories.set_index(column_labels['track'], inplace=True)
	trajectories = trajectories.groupby(column_labels['track'], group_keys=True).apply(lambda x: x.set_index(column_labels['time']).resample('1S').asfreq()).reset_index()
	trajectories[[column_labels['x'], column_labels['y']]] = trajectories.groupby(column_labels['track'], group_keys=False)[[column_labels['x'], column_labels['y']]].apply(lambda x: x.interpolate(method='linear'))
	trajectories.reset_index(drop=True, inplace=True)
	trajectories[column_labels['time']] = trajectories[column_labels['time']].astype('int64').astype(float) / 10**9
	#trajectories[column_labels['time']] = trajectories[column_labels['time']].astype('int64')
	trajectories.sort_values(by=[column_labels['track'],column_labels['time']],inplace=True)
	
	return trajectories
		

def extrapolate_tracks(trajectories, post=False, pre=False, column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):
	
	"""
	Extrapolate tracks in trajectories.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	post : bool, optional
		Flag indicating whether to perform post-extrapolation.
		Defaults to True.
	pre : bool, optional
		Flag indicating whether to perform pre-extrapolation.
		Defaults to False.
	column_labels : dict, optional
		Dictionary specifying the column labels used in the input DataFrame.
		The keys represent the following column labels:
		- 'track': The column label for the track ID.
		- 'time': The column label for the timestamp.
		- 'x': The column label for the x-coordinate.
		- 'y': The column label for the y-coordinate.
		Defaults to {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	pandas.DataFrame
		The extrapolated DataFrame with extended tracks.

	Notes
	-----
	This function extrapolates tracks in the input DataFrame by repeating the last known position
	either after (post-extrapolation) or before (pre-extrapolation) the available data.

	The input DataFrame is expected to have the following columns:
	- track: The unique ID of each track.
	- time: The timestamp of each data point.
	- x: The x-coordinate of each data point.
	- y: The y-coordinate of each data point.

	Examples
	--------
	>>> extrapolated_data = extrapolate_tracks(trajectories, post=True, pre=False, column_labels={'track': "ID", 'time': 'TIME', 'x': 'X', 'y': 'Y'})
	>>> print(extrapolated_data.head())

	"""
	
	if post:
		
		# get the maximum time T in the dataframe
		max_time = trajectories[column_labels['time']].max()

		# extrapolate the position until time T by repeating the last known position
		df_extrapolated = pd.DataFrame()
		for track_id, group in trajectories.groupby(column_labels['track']):
			last_known_position = group.loc[group[column_labels['time']] <= max_time].tail(1)[[column_labels['time'],column_labels['x'], column_labels['y']]].values
			extrapolated_frames = pd.DataFrame({column_labels['time']: np.arange(last_known_position[0][0] + 1, max_time + 1)})
			extrapolated_positions = pd.DataFrame({column_labels['x']: last_known_position[0][1], column_labels['y']: last_known_position[0][2]}, index=np.arange(last_known_position[0][0] + 1, max_time + 1))
			track_data = extrapolated_frames.join(extrapolated_positions, how="inner", on=column_labels['time'])
			track_data[column_labels['track']] = track_id
			df_extrapolated = pd.concat([df_extrapolated, track_data])


		# concatenate the original dataframe and the extrapolated dataframe
		trajectories = pd.concat([trajectories, df_extrapolated], axis=0)
		# sort the dataframe by TRACK_ID and FRAME
		trajectories.sort_values([column_labels['track'], column_labels['time']], inplace=True)

	if pre:
		
		# get the maximum time T in the dataframe
		min_time = 0 #trajectories[column_labels['time']].min()

		# extrapolate the position until time T by repeating the last known position
		df_extrapolated = pd.DataFrame()
		for track_id, group in trajectories.groupby(column_labels['track']):
			last_known_position = group.loc[group[column_labels['time']] >= min_time].head(1)[[column_labels['time'],column_labels['x'], column_labels['y']]].values
			extrapolated_frames = pd.DataFrame({column_labels['time']: np.arange(min_time, last_known_position[0][0] + 1)})
			extrapolated_positions = pd.DataFrame({column_labels['x']: last_known_position[0][1], column_labels['y']: last_known_position[0][2]}, index=np.arange(min_time, last_known_position[0][0]))
			track_data = extrapolated_frames.join(extrapolated_positions, how="inner", on=column_labels['time'])
			track_data[column_labels['track']] = track_id
			df_extrapolated = pd.concat([df_extrapolated, track_data])

		# concatenate the original dataframe and the extrapolated dataframe
		trajectories = pd.concat([trajectories, df_extrapolated], axis=0)

		# sort the dataframe by TRACK_ID and FRAME
		trajectories.sort_values([column_labels['track'], column_labels['time']], inplace=True)
	
	return trajectories

def compute_instantaneous_velocity(trajectories, column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):

	"""
	
	Compute the instantaneous velocity for each point in the trajectories.

	Parameters
	----------
	trajectories : pandas.DataFrame
		The input DataFrame containing trajectory data.
	column_labels : dict, optional
		A dictionary specifying the column labels for track ID, time, position X, and position Y.
		Defaults to {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	pandas.DataFrame
		The DataFrame with added 'velocity' column representing the instantaneous velocity for each point.

	Notes
	-----
	This function calculates the instantaneous velocity for each point in the trajectories.
	The velocity is computed as the Euclidean distance traveled divided by the time difference between consecutive points.

	The input DataFrame is expected to have columns with the specified column labels for track ID, time, position X, and position Y.

	Examples
	--------
	>>> velocity_data = compute_instantaneous_velocity(trajectories)
	>>> print(velocity_data.head())

	"""

	# Calculate the time differences and position differences
	trajectories['dt'] = trajectories.groupby(column_labels['track'])[column_labels['time']].diff()
	trajectories['dx'] = trajectories.groupby(column_labels['track'])[column_labels['x']].diff()
	trajectories['dy'] = trajectories.groupby(column_labels['track'])[column_labels['y']].diff()

	# Calculate the instantaneous velocity
	trajectories['velocity'] = np.sqrt(trajectories['dx']**2 +trajectories['dy']**2) / trajectories['dt']
	trajectories = trajectories.drop(['dx', 'dy', 'dt'], axis=1)
	trajectories = trajectories.sort_values(by=[column_labels['track'],column_labels['time']])    

	return trajectories

def instantaneous_diffusion(positions_x, positions_y, timeline):
	
	"""
	Compute the instantaneous diffusion coefficients for each position coordinate.

	Parameters
	----------
	positions_x : numpy.ndarray
		Array of x-coordinates of positions.
	positions_y : numpy.ndarray
		Array of y-coordinates of positions.
	timeline : numpy.ndarray
		Array of corresponding time points.

	Returns
	-------
	numpy.ndarray
		Array of instantaneous diffusion coefficients for each position coordinate.

	Notes
	-----
	The function calculates the instantaneous diffusion coefficients for each position coordinate (x, y) based on the provided positions and timeline.
	The diffusion coefficient at each time point is computed using the formula:
	D = ((x[t+1] - x[t-1])^2 / (2 * (t[t+1] - t[t-1]))) + (1 / (t[t+1] - t[t-1])) * ((x[t+1] - x[t]) * (x[t] - x[t-1]))
	where x represents the position coordinate (x or y) and t represents the corresponding time point.

	Examples
	--------
	>>> x = np.array([0, 1, 2, 3, 4, 5])
	>>> y = np.array([0, 1, 4, 9, 16, 25])
	>>> t = np.array([0, 1, 2, 3, 4, 5])
	>>> diff = instantaneous_diffusion(x, y, t)
	>>> print(diff)

	"""

	diff = np.zeros((len(positions_x),2))
	diff[:,:] = np.nan
	
	for t in range(1,len(positions_x)-1):
		diff[t,0] = (positions_x[t+1] - positions_x[t-1])**2/(2*(timeline[t+1] - timeline[t-1])) + 1/(timeline[t+1] - timeline[t-1])*((positions_x[t+1] - positions_x[t])*(positions_x[t] - positions_x[t-1]))

	for t in range(1,len(positions_y)-1):
		diff[t,1] = (positions_y[t+1] - positions_y[t-1])**2/(2*(timeline[t+1] - timeline[t-1])) + 1/(timeline[t+1] - timeline[t-1])*((positions_y[t+1] - positions_y[t])*(positions_y[t] - positions_y[t-1]))

	return diff

def magnitude_diffusion(diffusion_vector):

	"""
	Compute the magnitude of diffusion for each diffusion vector.

	Parameters
	----------
	diffusion_vector : numpy.ndarray
		Array of diffusion vectors.

	Returns
	-------
	numpy.ndarray
		Array of magnitudes of diffusion.

	Notes
	-----
	The function calculates the magnitude of diffusion for each diffusion vector (x, y) based on the provided diffusion vectors.
	The magnitude of diffusion is computed as the Euclidean norm of the diffusion vector.

	Examples
	--------
	>>> diffusion = np.array([[1.0, 2.0], [3.0, 4.0], [0.5, 0.5]])
	>>> magnitudes = magnitude_diffusion(diffusion)
	>>> print(magnitudes)

	"""

	return np.sqrt(diffusion_vector[:,0]**2+diffusion_vector[:,1]**2)


def compute_instantaneous_diffusion(trajectories, column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):

	"""
	
	Compute the instantaneous diffusion for each track in the provided trajectories DataFrame.

	Parameters
	----------
	trajectories : DataFrame
		The input DataFrame containing trajectories with position and time information.
	column_labels : dict, optional
		A dictionary specifying the column labels for track ID, time, x-coordinate, and y-coordinate. 
		The default is {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}.

	Returns
	-------
	DataFrame
		The modified DataFrame with an additional column "diffusion" containing the computed diffusion values.

	Notes
	-----

	The instantaneous diffusion is calculated using the positions and times of each track. The diffusion values
	are computed for each track individually and added as a new column "diffusion" in the output DataFrame.

	Examples
	--------
	>>> trajectories = pd.DataFrame({'TRACK_ID': [1, 1, 1, 2, 2, 2],
	...                              'FRAME': [0, 1, 2, 0, 1, 2],
	...                              'POSITION_X': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
	...                              'POSITION_Y': [0.5, 0.6, 0.7, 0.8, 0.9, 1.0]})
	>>> compute_instantaneous_diffusion(trajectories)
	# Output DataFrame with added "diffusion" column

	"""

	trajectories = trajectories.sort_values(by=[column_labels['track'],column_labels['time']])
	trajectories['diffusion'] = np.nan  
	
	for tid,group in trajectories.groupby(column_labels['track']):

		indices = group.index
		x = group[column_labels['x']].to_numpy()
		y = group[column_labels['y']].to_numpy()
		t = group[column_labels['time']].to_numpy()
		
		if len(x)>3: #to have t-1,t,t+1
			diff = instantaneous_diffusion(x,y,t)
			d = magnitude_diffusion(diff)
			trajectories.loc[indices, "diffusion"] = d

	return trajectories

def track_at_position(pos, mode, return_tracks=False, view_on_napari=False, threads=1):
	
	pos = pos.replace('\\','/')
	pos = rf"{pos}"
	assert os.path.exists(pos),f'Position {pos} is not a valid path.'
	if not pos.endswith('/'):
		pos += '/'

	script_path = os.sep.join([abs_path, 'scripts', 'track_cells.py'])	
	cmd = f'python "{script_path}" --pos "{pos}" --mode "{mode}" --threads "{threads}"'
	subprocess.call(cmd, shell=True)

	track_table = pos + os.sep.join(["output","tables",f"trajectories_{mode}.csv"])
	if return_tracks:
		df = pd.read_csv(track_table)
		return df
	else:
		return None	
	
	# # if return_labels or view_on_napari:
	# # 	labels = locate_labels(pos, population=mode)
	# # if view_on_napari:
	# # 	if stack_prefix is None:
	# # 		stack_prefix = ''
	# # 	stack = locate_stack(pos, prefix=stack_prefix)
	# # 	_view_on_napari(tracks=None, stack=stack, labels=labels)
	# # if return_labels:
	# # 	return labels
	# # else:
	# return None

def write_first_detection_class(tab, column_labels={'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}):
	
	"""
	Annotates a dataframe with the time of the first detection and classifies tracks based on their detection status.

	This function processes a dataframe containing tracking data, identifying the first point of detection for each
	track based on the x-coordinate values. It annotates the dataframe with the time of the first detection and
	assigns a class to each track indicating whether the first detection occurs at the start, during, or if there's
	no detection within the tracking data.

	Parameters
	----------
	tab : pandas.DataFrame
		The dataframe containing tracking data, expected to have columns for track ID, time, and spatial coordinates.
	column_labels : dict, optional
		A dictionary mapping standard column names ('track', 'time', 'x', 'y') to the corresponding column names in
		`tab`. Default column names are 'TRACK_ID', 'FRAME', 'POSITION_X', 'POSITION_Y'.

	Returns
	-------
	pandas.DataFrame
		The input dataframe `tab` with two additional columns: 'class_firstdetection' indicating the detection class,
		and 't_firstdetection' indicating the time of the first detection.

	Notes
	-----
	- Detection is based on the presence of non-NaN values in the 'x' column for each track.
	- Tracks with their first detection at the first time point are classified differently (`cclass=2`) and assigned
	  a `t_first` of -1, indicating no prior detection.
	- The function assumes uniform time steps between each frame in the tracking data.

	"""

	tab = tab.sort_values(by=[column_labels['track'],column_labels['time']])
	for tid,track_group in tab.groupby(column_labels['track']):
		indices = track_group.index
		detection = track_group[column_labels['x']].values
		timeline = track_group[column_labels['time']].values
		if len(timeline)>2:
			dt = timeline[1] - timeline[0]
			if np.any(detection==detection):
				t_first = timeline[detection==detection][0]
				cclass = 0
				if t_first==0:
					t_first = -1
					cclass = 2
				else:
					t_first =  float(t_first) - float(dt)
			else:
				t_first = -1
				cclass = 2

			tab.loc[indices, 'class_firstdetection'] = cclass
			tab.loc[indices, 't_firstdetection'] = t_first
	return tab



if __name__ == "__main__":
	track_at_position("/home/limozin/Documents/Experiments/MinimumJan/W4/401",
					  "targets",
					  )


from PyQt5.QtWidgets import QMainWindow, QApplication, QMessageBox, QScrollArea, QComboBox, QFrame, QCheckBox, QFileDialog, QGridLayout, QLineEdit, QVBoxLayout, QWidget, QLabel, QHBoxLayout, QPushButton
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QDoubleValidator, QIntValidator, QIcon
from celldetective.gui.gui_utils import center_window
from celldetective.gui.layouts import ChannelNormGenerator
from superqt import QLabeledDoubleSlider, QLabeledSlider, QSearchableComboBox
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
from celldetective.utils import get_software_location
from celldetective.io import locate_signal_dataset, get_signal_datasets_list, load_experiment_tables
from celldetective.signals import train_signal_model
import numpy as np
import json
import os
from glob import glob
from datetime import datetime
from celldetective.gui import Styles
from pandas.api.types import is_numeric_dtype

class ConfigSignalModelTraining(QMainWindow, Styles):
	
	"""
	UI to set measurement instructions.

	"""

	def __init__(self, parent_window=None, signal_mode='single-cells'):
		
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Train signal model")
		self.setWindowIcon(QIcon(os.sep.join(['celldetective','icons','mexican-hat.png'])))
		self.mode = self.parent_window.mode
		self.exp_dir = self.parent_window.exp_dir
		self.soft_path = get_software_location()
		self.pretrained_model = None 
		self.dataset_folder = None
		self.current_neighborhood = None
		self.reference_population = None
		self.neighbor_population = None
		self.signal_mode = signal_mode

		if self.signal_mode=='single-cells':
			self.signal_models_dir = self.soft_path+os.sep+os.sep.join(['celldetective','models','signal_detection'])
		elif self.signal_mode=='pairs':
			self.signal_models_dir = self.soft_path+os.sep+os.sep.join(['celldetective','models','pair_signal_detection'])
			self.mode = 'pairs'

		self.onlyFloat = QDoubleValidator()
		self.onlyInt = QIntValidator()

		self.screen_height = self.parent_window.parent_window.parent_window.screen_height
		center_window(self)

		self.setMinimumWidth(500)
		self.setMinimumHeight(int(0.3*self.screen_height))
		self.setMaximumHeight(int(0.8*self.screen_height))
		self.populate_widget()
		#self.load_previous_measurement_instructions()

	def populate_widget(self):

		"""
		Create the multibox design.

		"""
		
		# Create button widget and layout
		self.scroll_area = QScrollArea(self)
		self.button_widget = QWidget()
		self.main_layout = QVBoxLayout()
		self.button_widget.setLayout(self.main_layout)
		self.main_layout.setContentsMargins(30,30,30,30)

		# first frame for FEATURES
		self.model_frame = QFrame()
		self.model_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
		self.populate_model_frame()
		self.main_layout.addWidget(self.model_frame)

		self.data_frame = QFrame()
		self.data_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
		self.populate_data_frame()
		self.main_layout.addWidget(self.data_frame)

		self.hyper_frame = QFrame()
		self.hyper_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
		self.populate_hyper_frame()
		self.main_layout.addWidget(self.hyper_frame)

		self.submit_btn = QPushButton('Train')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.prep_model)
		self.main_layout.addWidget(self.submit_btn)
		self.submit_btn.setEnabled(False)

		#self.populate_left_panel()
		#grid.addLayout(self.left_side, 0, 0, 1, 1)
		self.button_widget.adjustSize()

		self.scroll_area.setAlignment(Qt.AlignCenter)
		self.scroll_area.setWidget(self.button_widget)
		self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
		self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
		self.scroll_area.setWidgetResizable(True)
		self.setCentralWidget(self.scroll_area)
		self.show()

		QApplication.processEvents()
		self.adjustScrollArea()

	def populate_hyper_frame(self):

		"""
		Add widgets and layout in the POST-PROCESSING frame.
		"""

		grid = QGridLayout(self.hyper_frame)
		grid.setContentsMargins(30,30,30,30)
		grid.setSpacing(30)

		self.hyper_lbl = QLabel("HYPERPARAMETERS")
		self.hyper_lbl.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")
		grid.addWidget(self.hyper_lbl, 0, 0, 1, 4, alignment=Qt.AlignCenter)
		self.generate_hyper_contents()
		grid.addWidget(self.ContentsHyper, 1, 0, 1, 4, alignment=Qt.AlignTop)

	def generate_hyper_contents(self):

		self.ContentsHyper = QFrame()
		layout = QVBoxLayout(self.ContentsHyper)
		layout.setContentsMargins(0,0,0,0)

		lr_layout = QHBoxLayout()
		lr_layout.addWidget(QLabel('learning rate: '),30)
		self.lr_le = QLineEdit('0,01')
		self.lr_le.setValidator(self.onlyFloat)
		lr_layout.addWidget(self.lr_le, 70)
		layout.addLayout(lr_layout)

		bs_layout = QHBoxLayout()
		bs_layout.addWidget(QLabel('batch size: '),30)
		self.bs_le = QLineEdit('64')
		self.bs_le.setValidator(self.onlyInt)
		bs_layout.addWidget(self.bs_le, 70)
		layout.addLayout(bs_layout)

		epochs_layout = QHBoxLayout()
		epochs_layout.addWidget(QLabel('# epochs: '), 30)
		self.epochs_slider = QLabeledSlider()
		self.epochs_slider.setRange(1,3000)
		self.epochs_slider.setSingleStep(1)
		self.epochs_slider.setTickInterval(1)		
		self.epochs_slider.setOrientation(1)
		self.epochs_slider.setValue(300)
		epochs_layout.addWidget(self.epochs_slider, 70)
		layout.addLayout(epochs_layout)


	def populate_data_frame(self):

		"""
		Add widgets and layout in the POST-PROCESSING frame.
		"""

		grid = QGridLayout(self.data_frame)
		grid.setContentsMargins(30,30,30,30)
		grid.setSpacing(30)

		self.data_lbl = QLabel("DATA")
		self.data_lbl.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")
		grid.addWidget(self.data_lbl, 0, 0, 1, 4, alignment=Qt.AlignCenter)
		self.generate_data_contents()
		grid.addWidget(self.ContentsData, 1, 0, 1, 4, alignment=Qt.AlignTop)

	def populate_model_frame(self):

		"""
		Add widgets and layout in the FEATURES frame.
		"""

		grid = QGridLayout(self.model_frame)
		grid.setContentsMargins(30,30,30,30)
		grid.setSpacing(30)

		self.model_lbl = QLabel("MODEL")
		self.model_lbl.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")
		grid.addWidget(self.model_lbl, 0, 0, 1, 4, alignment=Qt.AlignCenter)

		self.generate_model_panel_contents()
		grid.addWidget(self.ContentsModel, 1, 0, 1, 4, alignment=Qt.AlignTop)


	def generate_data_contents(self):

		self.ContentsData = QFrame()
		layout = QVBoxLayout(self.ContentsData)
		layout.setContentsMargins(0,0,0,0)

		train_data_layout = QHBoxLayout()
		train_data_layout.addWidget(QLabel('Training data: '), 30)
		self.select_data_folder_btn = QPushButton('Choose folder')
		self.select_data_folder_btn.clicked.connect(self.showDialog_dataset)
		self.data_folder_label = QLabel('No folder chosen')
		train_data_layout.addWidget(self.select_data_folder_btn, 35)
		train_data_layout.addWidget(self.data_folder_label, 30)

		self.cancel_dataset = QPushButton()
		self.cancel_dataset.setIcon(icon(MDI6.close,color="black"))
		self.cancel_dataset.clicked.connect(self.clear_dataset)
		self.cancel_dataset.setStyleSheet(self.button_select_all)
		self.cancel_dataset.setIconSize(QSize(20, 20))
		self.cancel_dataset.setVisible(False)
		train_data_layout.addWidget(self.cancel_dataset, 5)


		layout.addLayout(train_data_layout)

		include_dataset_layout = QHBoxLayout()
		include_dataset_layout.addWidget(QLabel('include dataset: '),30)
		self.dataset_cb = QComboBox()

		available_datasets, self.datasets_path = get_signal_datasets_list(return_path=True)
		signal_datasets = ['--'] + available_datasets

		self.dataset_cb.addItems(signal_datasets)
		include_dataset_layout.addWidget(self.dataset_cb, 70)
		layout.addLayout(include_dataset_layout)

		augmentation_hbox = QHBoxLayout()
		augmentation_hbox.addWidget(QLabel('augmentation\nfactor: '), 30)
		self.augmentation_slider = QLabeledDoubleSlider()
		self.augmentation_slider.setSingleStep(0.01)
		self.augmentation_slider.setTickInterval(0.01)		
		self.augmentation_slider.setOrientation(1)
		self.augmentation_slider.setRange(1, 5)
		self.augmentation_slider.setValue(2)

		augmentation_hbox.addWidget(self.augmentation_slider, 70)
		layout.addLayout(augmentation_hbox)

		validation_split_layout = QHBoxLayout()
		validation_split_layout.addWidget(QLabel('validation split: '),30)
		self.validation_slider = QLabeledDoubleSlider()
		self.validation_slider.setSingleStep(0.01)
		self.validation_slider.setTickInterval(0.01)		
		self.validation_slider.setOrientation(1)
		self.validation_slider.setRange(0,0.9)
		self.validation_slider.setValue(0.25)		
		validation_split_layout.addWidget(self.validation_slider, 70)
		layout.addLayout(validation_split_layout)


	def generate_model_panel_contents(self):
		
		self.ContentsModel = QFrame()
		layout = QVBoxLayout(self.ContentsModel)
		layout.setContentsMargins(0,0,0,0)

		modelname_layout = QHBoxLayout()
		modelname_layout.addWidget(QLabel('Model name: '), 30)
		self.modelname_le = QLineEdit()
		self.modelname_le.setText(f"Untitled_model_{datetime.today().strftime('%Y-%m-%d')}")
		modelname_layout.addWidget(self.modelname_le, 70)
		layout.addLayout(modelname_layout)

		if self.signal_mode=='pairs':
			neighborhood_layout = QHBoxLayout()
			neighborhood_layout.addWidget(QLabel('neighborhood of interest: '), 30)
			self.neighborhood_choice_cb = QSearchableComboBox()
			self.fill_available_neighborhoods()
			neighborhood_layout.addWidget(self.neighborhood_choice_cb, 70)
			layout.addLayout(neighborhood_layout)			

		classname_layout = QHBoxLayout()
		classname_layout.addWidget(QLabel('event name: '), 30)
		self.class_name_le = QLineEdit()
		self.class_name_le.setText("")
		classname_layout.addWidget(self.class_name_le, 70)
		layout.addLayout(classname_layout)

		pretrained_layout = QHBoxLayout()
		pretrained_layout.setContentsMargins(0,0,0,0)
		pretrained_layout.addWidget(QLabel('Pretrained model: '), 30)

		self.browse_pretrained_btn = QPushButton('Choose folder')
		self.browse_pretrained_btn.clicked.connect(self.showDialog_pretrained)
		pretrained_layout.addWidget(self.browse_pretrained_btn, 35)

		self.pretrained_lbl = QLabel('No folder chosen')
		pretrained_layout.addWidget(self.pretrained_lbl, 30)

		self.cancel_pretrained = QPushButton()
		self.cancel_pretrained.setIcon(icon(MDI6.close,color="black"))
		self.cancel_pretrained.clicked.connect(self.clear_pretrained)
		self.cancel_pretrained.setStyleSheet(self.button_select_all)
		self.cancel_pretrained.setIconSize(QSize(20, 20))
		self.cancel_pretrained.setVisible(False)
		pretrained_layout.addWidget(self.cancel_pretrained, 5)

		layout.addLayout(pretrained_layout)

		recompile_layout = QHBoxLayout()
		recompile_layout.addWidget(QLabel('Recompile: '), 30)
		self.recompile_option = QCheckBox()
		self.recompile_option.setEnabled(False)
		recompile_layout.addWidget(self.recompile_option, 70)
		layout.addLayout(recompile_layout)

		self.max_nbr_channels = 5
		self.ch_norm = ChannelNormGenerator(self, mode='signals')
		layout.addLayout(self.ch_norm)

		if self.signal_mode=='pairs':
			self.neighborhood_choice_cb.currentIndexChanged.connect(self.neighborhood_changed)
			self.neighborhood_changed()

		model_length_layout = QHBoxLayout()
		model_length_layout.addWidget(QLabel('Max signal length: '), 30)
		self.model_length_slider = QLabeledSlider()
		self.model_length_slider.setSingleStep(1)
		self.model_length_slider.setTickInterval(1)
		self.model_length_slider.setSingleStep(1)
		self.model_length_slider.setOrientation(1)
		self.model_length_slider.setRange(0,1024)
		self.model_length_slider.setValue(128)		
		model_length_layout.addWidget(self.model_length_slider, 70)
		layout.addLayout(model_length_layout)

	def neighborhood_changed(self):

		neigh = self.neighborhood_choice_cb.currentText()
		self.current_neighborhood = neigh.replace('target_ref_','').replace('effector_ref_','')
		self.reference_population = ['targets' if 'target' in neigh else 'effectors'][0]
		if 'target' in neigh:
			if 'self' in neigh:
				self.neighbor_population = 'targets'
			else:
				self.neighbor_population = 'effectors'
		else:
			if 'self' in neigh:
				self.neighbor_population = 'effectors'
			else:
				self.neighbor_population = 'targets'
		
		print(f'Current neighborhood: {self.current_neighborhood}')
		print(f'New reference population: {self.reference_population}')
		print(f'New neighbor population: {self.neighbor_population}')

		# reload reference signals / neighbor signals / pair signals
		# fill the channel cbs
		self.df_reference = self.dataframes[self.reference_population]
		self.df_neighbor = self.dataframes[self.neighbor_population]
		self.df_pairs = load_experiment_tables(self.parent_window.exp_dir, population='pairs', load_pickle=False)

		self.df_reference = self.df_reference.rename(columns=lambda x: 'reference_' + x)
		num_cols_reference = [c for c in list(self.df_reference.columns) if is_numeric_dtype(self.df_reference[c])]
		self.df_neighbor = self.df_neighbor.rename(columns=lambda x: 'neighbor_' + x)
		num_cols_neighbor = [c for c in list(self.df_neighbor.columns) if is_numeric_dtype(self.df_neighbor[c])]
		self.df_pairs = self.df_pairs.rename(columns=lambda x: 'pair_' + x)
		num_cols_pairs = [c for c in list(self.df_pairs.columns) if is_numeric_dtype(self.df_pairs[c])]

		self.signals = ['--'] + num_cols_pairs + num_cols_reference + num_cols_neighbor

		for cb in self.ch_norm.channel_cbs:
			# try:
			# 	cb.disconnect()
			# except:
			# 	pass
			cb.clear()
			cb.addItems(self.signals)

	def fill_available_neighborhoods(self):
		
		df_targets = load_experiment_tables(self.parent_window.exp_dir, population='targets', load_pickle=True)
		df_effectors = load_experiment_tables(self.parent_window.exp_dir, population='effectors', load_pickle=True)

		self.dataframes = {
			'targets': df_targets,
			'effectors': df_effectors,
		}

		self.neighborhood_cols = []
		self.reference_populations = []
		self.neighbor_populations = []
		if df_targets is not None:
			self.neighborhood_cols.extend(['target_ref_'+c for c in list(df_targets.columns) if c.startswith('neighborhood')])
			self.reference_populations.extend(['targets' for c in list(df_targets.columns) if c.startswith('neighborhood')])
			for c in list(df_targets.columns):
				if c.startswith('neighborhood') and '_2_' in c:
					self.neighbor_populations.append('effectors')
				elif c.startswith('neighborhood') and 'self' in c:
					self.neighbor_populations.append('targets')
		
		if df_effectors is not None:
			self.neighborhood_cols.extend(['effector_ref_'+c for c in list(df_effectors.columns) if c.startswith('neighborhood')])
			self.reference_populations.extend(['effectors' for c in list(df_effectors.columns) if c.startswith('neighborhood')])
			for c in list(df_effectors.columns):
				if c.startswith('neighborhood') and '_2_' in c:
					self.neighbor_populations.append('targets')
				elif c.startswith('neighborhood') and 'self' in c:
					self.neighbor_populations.append('effectors')

		print(f"The following neighborhoods were detected: {self.neighborhood_cols=} {self.reference_populations=} {self.neighbor_populations=}")

		self.neighborhood_choice_cb.addItems(self.neighborhood_cols)

	def showDialog_pretrained(self):

		self.pretrained_model = QFileDialog.getExistingDirectory(
						self, "Open Directory",
						os.sep.join([self.soft_path,'celldetective','models','signal_detection','']),
						QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
						)

		if self.pretrained_model is not None:
		# 	self.foldername = self.file_dialog_pretrained.selectedFiles()[0]
			subfiles = glob(os.sep.join([self.pretrained_model,"*"]))
			if os.sep.join([self.pretrained_model,"config_input.json"]) in subfiles:
				self.load_pretrained_config()
				self.pretrained_lbl.setText(self.pretrained_model.split(os.sep)[-1])
				self.cancel_pretrained.setVisible(True)
				self.recompile_option.setEnabled(True)
				self.modelname_le.setText(f"{self.pretrained_model.split(os.sep)[-1]}_{datetime.today().strftime('%Y-%m-%d')}")
			else:
				self.pretrained_model = None
				self.pretrained_lbl.setText('No folder chosen')	
				self.recompile_option.setEnabled(False)	
				self.cancel_pretrained.setVisible(False)
		print(self.pretrained_model)

	def showDialog_dataset(self):

		self.dataset_folder = QFileDialog.getExistingDirectory(
						self, "Open Directory",
						self.exp_dir,
						QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
						)
		if self.dataset_folder is not None:

			subfiles = glob(os.sep.join([self.dataset_folder,"*.npy"]))
			if len(subfiles)>0:
				print(f'found {len(subfiles)} files in folder')
				self.data_folder_label.setText(self.dataset_folder[:16]+'...')
				self.data_folder_label.setToolTip(self.dataset_folder)
				self.cancel_dataset.setVisible(True)
			else:
				self.data_folder_label.setText('No folder chosen')
				self.data_folder_label.setToolTip('')
				self.dataset_folder = None
				self.cancel_dataset.setVisible(False)

	def clear_pretrained(self):
		
		self.pretrained_model = None
		self.pretrained_lbl.setText('No folder chosen')
		for cb in self.ch_norm.channel_cbs:
			cb.setEnabled(True)
		self.ch_norm.add_col_btn.setEnabled(True)
		self.recompile_option.setEnabled(False)
		self.cancel_pretrained.setVisible(False)
		self.model_length_slider.setEnabled(True)
		self.class_name_le.setText('')
		self.modelname_le.setText(f"Untitled_model_{datetime.today().strftime('%Y-%m-%d')}")

	def clear_dataset(self):

		self.dataset_folder = None
		self.data_folder_label.setText('No folder chosen')
		self.data_folder_label.setToolTip('')
		self.cancel_dataset.setVisible(False)


	def load_pretrained_config(self):

		f = open(os.sep.join([self.pretrained_model,"config_input.json"]))
		data = json.load(f)
		channels = data["channels"]
		signal_length = data["model_signal_length"]
		try:
			label = data['label']
			self.class_name_le.setText(label)
		except:
			pass
		self.model_length_slider.setValue(int(signal_length))
		self.model_length_slider.setEnabled(False)

		for c,cb in zip(channels, self.ch_norm.channel_cbs):
			index = cb.findText(c)
			cb.setCurrentIndex(index)

		if len(channels)<len(self.ch_norm.channel_cbs):
			for k in range(len(self.ch_norm.channel_cbs)-len(channels)):
				self.ch_norm.channel_cbs[len(channels)+k].setCurrentIndex(0)
				self.ch_norm.channel_cbs[len(channels)+k].setEnabled(False)
		self.ch_norm.add_col_btn.setEnabled(False)


	def adjustScrollArea(self):
		
		"""
		Auto-adjust scroll area to fill space 
		(from https://stackoverflow.com/questions/66417576/make-qscrollarea-use-all-available-space-of-qmainwindow-height-axis)
		"""

		step = 5
		while self.scroll_area.verticalScrollBar().isVisible() and self.height() < self.maximumHeight():
			self.resize(self.width(), self.height() + step)

	def prep_model(self):

		model_name = self.modelname_le.text()
		pretrained_model = self.pretrained_model
		signal_length = self.model_length_slider.value()
		recompile_op = self.recompile_option.isChecked()

		channels = []
		for i in range(len(self.ch_norm.channel_cbs)):
			channels.append(self.ch_norm.channel_cbs[i].currentText())

		slots_to_keep = np.where(np.array(channels)!='--')[0]
		while '--' in channels:
			channels.remove('--')

		norm_values = np.array([[float(a.replace(',','.')),float(b.replace(',','.'))] for a,b in zip([l.text() for l in self.ch_norm.normalization_min_value_le],
											[l.text() for l in self.ch_norm.normalization_max_value_le])])
		norm_values = norm_values[slots_to_keep]
		norm_values = [list(v) for v in norm_values]

		clip_values = np.array(self.ch_norm.clip_option)
		clip_values = list(clip_values[slots_to_keep])
		clip_values = [bool(c) for c in clip_values]

		normalization_mode = np.array(self.ch_norm.normalization_mode)
		normalization_mode = list(normalization_mode[slots_to_keep])
		normalization_mode = [bool(m) for m in normalization_mode]

		data_folders = []
		if self.dataset_folder is not None:
			data_folders.append(self.dataset_folder)
		if self.dataset_cb.currentText()!='--':
			dataset = locate_signal_dataset(self.dataset_cb.currentText())
			data_folders.append(dataset)

		aug_factor = self.augmentation_slider.value()
		val_split = self.validation_slider.value()

		try:
			lr = float(self.lr_le.text().replace(',','.'))
		except:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("Invalid value encountered for the learning rate.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None			
		
		bs = int(self.bs_le.text())
		epochs = self.epochs_slider.value()

		training_instructions = {'model_name': model_name,'pretrained': pretrained_model, 'channel_option': channels, 'normalization_percentile': normalization_mode,
		'normalization_clip': clip_values,'normalization_values': norm_values, 'model_signal_length': signal_length,
		'recompile_pretrained': recompile_op, 'ds': data_folders, 'augmentation_factor': aug_factor, 'validation_split': val_split,
		'learning_rate': lr, 'batch_size': bs, 'epochs': epochs, 'label': self.class_name_le.text(), 'neighborhood_of_interest': self.current_neighborhood, 'reference_population': self.reference_population, 'neighbor_population': self.neighbor_population}

		model_folder = self.signal_models_dir +os.sep+ model_name + os.sep
		if not os.path.exists(model_folder):
			os.mkdir(model_folder)

		training_instructions.update({'target_directory': self.signal_models_dir})

		print(f"Set of instructions: {training_instructions}")
		with open(model_folder+"training_instructions.json", 'w') as f:
			json.dump(training_instructions, f, indent=4)
		
		train_signal_model(model_folder+"training_instructions.json")

		self.parent_window.refresh_signal_models()